import matplotlib.pyplot as plt
from pandas import DataFrame, Series, DatetimeIndex

from archetypal import EnergySeries, settings
from archetypal.energyseries import plot_energyseries_map, save_and_show


class EnergyDataFrame(DataFrame):
    """An EnergyDataFrame object is a pandas.DataFrame that has energy related
    data. In addition to the standard DataFrame constructor arguments,
    EnergyDataFrame also accepts the following keyword arguments:


    """

    _metadata = [
        "profile_type",
        "base_year",
        "frequency",
        "units",
        "sort_values",
        "to_units",
    ]

    def __init__(self, *args, **kwargs):
        from_units = kwargs.pop("units", None)
        super(EnergyDataFrame, self).__init__(*args, **kwargs)
        self.from_units = from_units
        if from_units is not None:
            self.set_unit(from_units, inplace=True)

    def set_unit(self, from_unit, inplace):
        ureg = settings.unit_registry

        if inplace:
            frame = self
        else:
            frame = self.copy()

        self.from_units = ureg.parse_expression(from_unit)

        if not inplace:
            return frame

    def plot2d(self, **kwargs):
        return plot_energydataframe_map(self, **kwargs)

    @property
    def _constructor(self):
        return EnergyDataFrame

    @property
    def nseries(self):
        if self._data.ndim == 1:
            return 1
        else:
            return self._data.shape[0]

    def __getitem__(self, key):
        """
        return an EnergySeries or an EnergyDataFrame with propagated metadata.
        """
        result = super(EnergyDataFrame, self).__getitem__(key)
        if isinstance(result, Series):
            result.__class__ = EnergySeries
        elif isinstance(result, DataFrame):
            result.__class__ = EnergyDataFrame
        return result.__finalize__(self)

    def stack(self, level=-1, dropna=True):
        from pandas.core.reshape.reshape import stack, stack_multiple

        if isinstance(level, (tuple, list)):
            result = stack_multiple(self, level, dropna=dropna)
            return result.__finalize__(self)
        else:
            result = stack(self, level, dropna=dropna)
            return result.__finalize__(self)

    def discretize_tsam(
        self,
        resolution=None,
        noTypicalPeriods=10,
        hoursPerPeriod=24,
        clusterMethod="hierarchical",
        evalSumPeriods=False,
        sortValues=False,
        sameMean=False,
        rescaleClusterPeriods=True,
        weightDict=None,
        extremePeriodMethod="None",
        solver="glpk",
        roundOutput=None,
        addPeakMin=None,
        addPeakMax=None,
        addMeanMin=None,
        addMeanMax=None,
    ):
        """uses tsam"""
        try:
            import tsam.timeseriesaggregation as tsam
        except ImportError:
            raise ImportError("tsam is required for discretize_tsam()")
        if not isinstance(self.index, DatetimeIndex):
            raise TypeError("To use tsam, index of series must be a " "DateTimeIndex")
        timeSeries = self.copy()
        agg = tsam.TimeSeriesAggregation(
            timeSeries,
            resolution=resolution,
            noTypicalPeriods=noTypicalPeriods,
            hoursPerPeriod=hoursPerPeriod,
            clusterMethod=clusterMethod,
            evalSumPeriods=evalSumPeriods,
            sortValues=sortValues,
            sameMean=sameMean,
            rescaleClusterPeriods=rescaleClusterPeriods,
            weightDict=weightDict,
            extremePeriodMethod=extremePeriodMethod,
            solver=solver,
            roundOutput=roundOutput,
            addPeakMin=addPeakMin,
            addPeakMax=addPeakMax,
            addMeanMin=addMeanMin,
            addMeanMax=addMeanMax,
        )

        agg.createTypicalPeriods()
        results = agg.predictOriginalData()
        results = EnergyDataFrame(results)
        results.__dict__["agg"] = agg
        return results.__finalize__(self)


def plot_energydataframe_map(
    data,
    periodlength=24,
    subplots=False,
    vmin=None,
    vmax=None,
    axis_off=True,
    cmap="RdBu",
    fig_height=None,
    fig_width=6,
    show=True,
    view_angle=-60,
    save=False,
    close=False,
    dpi=300,
    file_format="png",
    color=None,
    ax=None,
    filename="untitled",
    extent="tight",
    sharex=True,
    sharey=True,
    layout=None,
    layout_type="vertical",
    **kwargs
):
    if fig_height is None:
        fig_height = fig_width / 3
    figsize = (fig_width, fig_height)
    nseries = data.nseries
    fig, axes = _setup_subplots(
        subplots, nseries, sharex, sharey, figsize, ax, layout, layout_type
    )
    cols = data.columns
    for ax, col in zip(axes, cols):
        plot_energyseries_map(
            data[col],
            periodlength=periodlength,
            subplots=subplots,
            vmin=vmin,
            vmax=vmax,
            axis_off=axis_off,
            cmap=cmap,
            fig_height=fig_height,
            fig_width=fig_width,
            show=False,
            save=False,
            close=False,
            dpi=dpi,
            file_format=file_format,
            color=color,
            ax=ax,
            filename=filename,
            extent=extent,
            sharex=sharex,
            sharey=sharey,
            layout=layout,
            layout_type=layout_type,
            **kwargs
        )

    fig, axes = save_and_show(
        fig, axes, save, show, close, filename, file_format, dpi, axis_off, extent
    )

    return fig, axes


def _setup_subplots(
    subplots,
    nseries,
    sharex=False,
    sharey=False,
    figsize=None,
    ax=None,
    layout=None,
    layout_type="vertical",
):
    """prepares the subplots"""
    from pandas.plotting._tools import _subplots, _flatten

    if subplots:
        fig, axes = _subplots(
            naxes=nseries,
            sharex=sharex,
            sharey=sharey,
            figsize=figsize,
            ax=ax,
            layout=layout,
            layout_type=layout_type,
        )
    else:
        if ax is None:
            fig = plt.figure(figsize=figsize)
            axes = fig.add_subplot(111)
        else:
            fig = ax.get_figure()
            if figsize is not None:
                fig.set_size_inches(figsize)
            axes = ax

    axes = _flatten(axes)

    return fig, axes



################################################################################
# Module: archetypal.template
# Description:
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import collections

from archetypal import log, timeit, settings
from archetypal.template import UmiBase, Unique, UmiSchedule, UniqueName
from archetypal.utils import reduce
import logging as lg


class ZoneLoad(UmiBase, metaclass=Unique):
    """Zone Loads

    Important:
        Please note that the calculation of the equipment power density will sum
        up the electric equipment objects as well as the gas equipment objects.

    .. image:: ../images/template/zoneinfo-loads.png
    """

    def __init__(
        self,
        DimmingType="Continuous",
        EquipmentAvailabilitySchedule=None,
        EquipmentPowerDensity=12,
        IlluminanceTarget=500,
        LightingPowerDensity=12,
        LightsAvailabilitySchedule=None,
        OccupancySchedule=None,
        IsEquipmentOn=True,
        IsLightingOn=True,
        IsPeopleOn=True,
        PeopleDensity=0.2,
        **kwargs,
    ):
        """Initialize a new ZoneLoad object

        Args:
            DimmingType (str): Different types to dim the lighting to respect the
                IlluminanceTraget and taking into account the daylight illuminance:
                    - If `Continuous`, the overhead lights dim continuously and linearly
                      from (maximum electric power, maximum light output) to (minimum
                      electric power, minimum light output) as the daylight illuminance
                      increases. The lights stay on at the minimum point with further
                      increase in the daylight illuminance.
                    - If `Stepped`, the electric power input and light output vary
                      in discrete, equally spaced steps.
                    - If `Off`, Lights switch off completely when the minimum
                      dimming point is reached.
            EquipmentAvailabilitySchedule (UmiSchedule): The name of
                the schedule (Day | Week | Year) that modifies the design level
                parameter for electric equipment.
            EquipmentPowerDensity (float): Equipment Power Density in the zone
                (W/m²).
            IlluminanceTarget (float): Number of lux to be respected in the zone
            LightingPowerDensity (float): Lighting Power Density in the zone
                (W/m²).
            LightsAvailabilitySchedule (UmiSchedule): The name of the
                schedule (Day | Week | Year) that modifies the design level
                parameter for lighting.
            OccupancySchedule (UmiSchedule): The name of the schedule
                (Day | Week | Year) that modifies the number of people parameter
                for electric equipment.
            IsEquipmentOn (bool): If True, heat gains from Equipment are taken
                into account for the zone's load calculation.
            IsLightingOn (bool): If True, heat gains from Lights are taken into
                account for the zone's load calculation.
            IsPeopleOn (bool): If True, heat gains from People are taken into
                account for the zone's load calculation.
            PeopleDensity (float): Density of people in the zone (people/m²).
            **kwargs: Other keywords passed to the parent constructor :class:`UmiBase`.
        """
        super(ZoneLoad, self).__init__(**kwargs)
        self.DimmingType = DimmingType
        self.EquipmentAvailabilitySchedule = EquipmentAvailabilitySchedule
        self.EquipmentPowerDensity = EquipmentPowerDensity
        self.IlluminanceTarget = IlluminanceTarget
        self.LightingPowerDensity = LightingPowerDensity
        self.LightsAvailabilitySchedule = LightsAvailabilitySchedule
        self.OccupancySchedule = OccupancySchedule
        self.IsEquipmentOn = IsEquipmentOn
        self.IsLightingOn = IsLightingOn
        self.IsPeopleOn = IsPeopleOn
        self.PeopleDensity = PeopleDensity

        self._belongs_to_zone = kwargs.get("zone", None)

    def __add__(self, other):
        """
        Args:
            other (Zone):
        """
        return self.combine(other)

    def __hash__(self):
        return hash((self.__class__.__name__, self.Name, self.DataSource))

    def __eq__(self, other):
        if not isinstance(other, ZoneLoad):
            return False
        else:
            return all(
                [
                    self.DimmingType == other.DimmingType,
                    self.EquipmentAvailabilitySchedule
                    == other.EquipmentAvailabilitySchedule,
                    self.EquipmentPowerDensity == other.EquipmentPowerDensity,
                    self.IlluminanceTarget == other.IlluminanceTarget,
                    self.LightingPowerDensity == other.LightingPowerDensity,
                    self.LightsAvailabilitySchedule == other.LightsAvailabilitySchedule,
                    self.OccupancySchedule == other.OccupancySchedule,
                    self.IsEquipmentOn == other.IsEquipmentOn,
                    self.IsLightingOn == other.IsLightingOn,
                    self.IsPeopleOn == other.IsPeopleOn,
                    self.PeopleDensity == other.PeopleDensity,
                ]
            )

    @classmethod
    def from_json(cls, *args, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        zl = cls(*args, **kwargs)

        cool_schd = kwargs.get("EquipmentAvailabilitySchedule", None)
        zl.EquipmentAvailabilitySchedule = zl.get_ref(cool_schd)
        heat_schd = kwargs.get("LightsAvailabilitySchedule", None)
        zl.LightsAvailabilitySchedule = zl.get_ref(heat_schd)
        mech_schd = kwargs.get("OccupancySchedule", None)
        zl.OccupancySchedule = zl.get_ref(mech_schd)

        return zl

    def to_json(self):
        """Convert class properties to dict"""
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["DimmingType"] = self.DimmingType
        data_dict[
            "EquipmentAvailabilitySchedule"
        ] = self.EquipmentAvailabilitySchedule.to_dict()
        data_dict["EquipmentPowerDensity"] = self.EquipmentPowerDensity
        data_dict["IlluminanceTarget"] = self.IlluminanceTarget
        data_dict["LightingPowerDensity"] = self.LightingPowerDensity
        data_dict[
            "LightsAvailabilitySchedule"
        ] = self.LightsAvailabilitySchedule.to_dict()
        data_dict["OccupancySchedule"] = self.OccupancySchedule.to_dict()
        data_dict["IsEquipmentOn"] = self.IsEquipmentOn
        data_dict["IsLightingOn"] = self.IsLightingOn
        data_dict["IsPeopleOn"] = self.IsPeopleOn
        data_dict["PeopleDensity"] = self.PeopleDensity
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    @classmethod
    @timeit
    def from_zone(cls, zone):
        """
        Args:
            zone (archetypal.template.zone.Zone): zone to gets information from
        """

        # Get schedule index for different loads and create ZoneLoad arguments
        # Verify if Equipment in zone
        zone_index = zone.sql["Zones"][
            zone.sql["Zones"]["ZoneName"].str.contains(zone.Name.upper())
        ].index[0]
        nominal_elec = zone.sql["NominalElectricEquipment"][
            zone.sql["NominalElectricEquipment"]["ZoneIndex"] == zone_index
        ]
        nominal_gas = zone.sql["NominalGasEquipment"][
            zone.sql["NominalGasEquipment"]["ZoneIndex"] == zone_index
        ]
        if nominal_elec.empty and nominal_gas.empty:
            EquipmentAvailabilitySchedule = UmiSchedule.constant_schedule(idf=zone.idf)
            EquipmentPowerDensity = 0.0
        else:
            if nominal_gas.empty:
                EquipmentPowerDensity = nominal_elec["DesignLevel"].sum() / zone.area
            elif nominal_elec.empty:
                EquipmentPowerDensity = nominal_gas["DesignLevel"].sum() / zone.area
            else:
                EquipmentPowerDensity = (
                    nominal_elec["DesignLevel"].sum() + nominal_gas["DesignLevel"].sum()
                ) / zone.area  # todo: Should nominal gas really be added to elec?

            sched_indexes = nominal_elec["ScheduleIndex"].values
            design_index = nominal_elec["DesignLevel"].index
            list_sched = []
            for sched, design in zip(sched_indexes, design_index):
                sched_name = zone.sql["Schedules"]["ScheduleName"][sched]
                schedule = UmiSchedule(Name=sched_name, idf=zone.idf)
                schedule.combine_weight = nominal_elec["DesignLevel"][design]
                list_sched.append(schedule)

            EquipmentAvailabilitySchedule = reduce(
                UmiSchedule.combine, list_sched, weights="combine_weight"
            )
        # Verifies if Lights in zone
        if zone.sql["NominalLighting"][
            zone.sql["NominalLighting"]["ZoneIndex"] == zone_index
        ].empty:
            LightsAvailabilitySchedule = UmiSchedule.constant_schedule(idf=zone.idf)
            LightingPowerDensity = 0.0
        else:
            schedule_light_index = zone.sql["NominalLighting"][
                zone.sql["NominalLighting"]["ZoneIndex"] == zone_index
            ]["ScheduleIndex"].iloc[0]
            LightsAvailabilitySchedule = UmiSchedule(
                Name=zone.sql["Schedules"]["ScheduleName"].iloc[
                    schedule_light_index - 1
                ],
                idf=zone.idf,
            )
            LightingPowerDensity = (
                zone.sql["NominalLighting"][
                    zone.sql["NominalLighting"]["ZoneIndex"] == zone_index
                ]["DesignLevel"].iloc[0]
                / zone.area
            )
        # Verifies if People in zone
        if zone.sql["NominalPeople"][
            zone.sql["NominalPeople"]["ZoneIndex"] == zone_index
        ].empty:
            OccupancySchedule = UmiSchedule.constant_schedule(idf=zone.idf)
            PeopleDensity = 0.0
        else:
            schedule_people_index = zone.sql["NominalPeople"][
                zone.sql["NominalPeople"]["ZoneIndex"] == zone_index
            ]["NumberOfPeopleScheduleIndex"].iloc[0]
            OccupancySchedule = UmiSchedule(
                Name=zone.sql["Schedules"]["ScheduleName"].iloc[
                    schedule_people_index - 1
                ],
                idf=zone.idf,
            )
            PeopleDensity = (
                zone.sql["NominalPeople"][
                    zone.sql["NominalPeople"]["ZoneIndex"] == zone_index
                ]["NumberOfPeople"].iloc[0]
                / zone.area
            )

        name = zone.Name + "_ZoneLoad"
        z_load = cls(
            Name=name,
            zone=zone,
            DimmingType=_resolve_dimming_type(zone),
            EquipmentAvailabilitySchedule=EquipmentAvailabilitySchedule,
            EquipmentPowerDensity=EquipmentPowerDensity,
            IlluminanceTarget=_resolve_illuminance_target(zone),
            LightingPowerDensity=LightingPowerDensity,
            LightsAvailabilitySchedule=LightsAvailabilitySchedule,
            OccupancySchedule=OccupancySchedule,
            IsEquipmentOn=EquipmentPowerDensity > 0,
            IsLightingOn=LightingPowerDensity > 0,
            IsPeopleOn=PeopleDensity > 0,
            PeopleDensity=PeopleDensity,
            idf=zone.idf,
            Category=zone.idf.building_name(use_idfname=True),
        )
        return z_load

    def combine(self, other, weights=None):
        """Combine two ZoneLoad objects together.

        Args:
            other (ZoneLoad):
            weights (list-like, optional): A list-like object of len 2. If None,
                the volume of the zones for which self and other belongs is
                used.

        Returns:
            (ZoneLoad): the combined ZoneLoad object.
        """
        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        incoming_load_data = self.__dict__.copy()
        incoming_load_data.pop("Name")

        meta = self._get_predecessors_meta(other)

        if not weights:
            zone_weight = settings.zone_weight
            weights = [
                getattr(self._belongs_to_zone, str(zone_weight)),
                getattr(other._belongs_to_zone, str(zone_weight)),
            ]
            log(
                'using zone {} "{}" as weighting factor in "{}" '
                "combine.".format(
                    zone_weight,
                    " & ".join(list(map(str, map(int, weights)))),
                    self.__class__.__name__,
                )
            )

        attr = dict(
            DimmingType=self._str_mean(other, "DimmingType"),
            EquipmentAvailabilitySchedule=self.EquipmentAvailabilitySchedule.combine(
                other.EquipmentAvailabilitySchedule,
                weights=weights,
                quantity=[self.EquipmentPowerDensity, other.EquipmentPowerDensity],
            ),
            EquipmentPowerDensity=self._float_mean(
                other, "EquipmentPowerDensity", weights
            ),
            IlluminanceTarget=self._float_mean(other, "IlluminanceTarget", weights),
            LightingPowerDensity=self._float_mean(
                other, "LightingPowerDensity", weights
            ),
            LightsAvailabilitySchedule=self.LightsAvailabilitySchedule.combine(
                other.LightsAvailabilitySchedule,
                weights=weights,
                quantity=[self.LightingPowerDensity, other.LightingPowerDensity],
            ),
            OccupancySchedule=self.OccupancySchedule.combine(
                other.OccupancySchedule,
                weights=weights,
                quantity=[self.PeopleDensity, other.PeopleDensity],
            ),
            IsEquipmentOn=any([self.IsEquipmentOn, other.IsEquipmentOn]),
            IsLightingOn=any([self.IsLightingOn, other.IsLightingOn]),
            IsPeopleOn=any([self.IsPeopleOn, other.IsPeopleOn]),
            PeopleDensity=self._float_mean(other, "PeopleDensity", weights),
        )

        new_obj = self.__class__(**meta, **attr)
        new_obj._belongs_to_zone = self._belongs_to_zone
        new_obj._predecessors.extend(self.predecessors + other.predecessors)
        return new_obj


def _resolve_dimming_type(zone):
    """Resolves the dimming type for the Zone object"""
    # First, retrieve the list of Daylighting objects for this zone. Uses the eppy
    # `getreferingobjs` method.
    ep_obj = zone._epbunch
    possible_ctrls = ep_obj.getreferingobjs(
        iddgroups=["Daylighting"], fields=["Zone_Name"]
    )
    # Then, if there are controls
    if possible_ctrls:
        # Filter only the "Daylighting:Controls"
        ctrls = [
            ctrl
            for ctrl in possible_ctrls
            if ctrl.key.upper() == "Daylighting:Controls".upper()
        ]
        ctrl_types = [ctrl["Lighting_Control_Type"] for ctrl in ctrls]

        # There should only be one control per zone. A set of controls should return 1.
        if len(set(ctrl_types)) == 1:
            dimming_type = next(iter(set(ctrl_types)))
            if dimming_type.lower() not in ["continuous", "stepped"]:
                raise ValueError(
                    f"A dimming type of type '{dimming_type}' for zone '{zone.Name}' is not yet supported in UMI"
                )
            else:
                log(f"Dimming type for zone '{zone.Name}' set to '{dimming_type}'")
                return dimming_type  # Return first element
        else:
            raise ValueError(
                "Could not resolve more than one dimming types for Zone {}. "
                "Make sure there is only one".format(zone.Name)
            )
    else:
        # Else, there are no dimming controls => set to "Off".
        log(
            "No dimming type found for zone {}. Setting as Off".format(zone.Name),
            lg.DEBUG,
        )
        return "Off"


def _resolve_illuminance_target(zone):
    """Resolves the illuminance target for the Zone object"""
    # First, retrieve the list of Daylighting objects for this zone. Uses the eppy
    # `getreferingobjs` method.
    ep_obj = zone._epbunch
    possible_ctrls = ep_obj.getreferingobjs(
        iddgroups=["Daylighting"], fields=["Zone_Name"]
    )
    # Then, if there are controls
    if possible_ctrls:
        # Filter only the "Daylighting:Controls"
        ctrls = [
            ctrl
            for ctrl in possible_ctrls
            if ctrl.key.upper() == "Daylighting:Controls".upper()
        ]
        ctrl_types = [
            ctrl["Illuminance_Setpoint_at_Reference_Point_1"] for ctrl in ctrls
        ]

        # There should only be one control per zone. A set of controls should return 1.
        if len(set(ctrl_types)) == 1:
            dimming_type = next(iter(set(ctrl_types)))
            log(f"Illuminance target for zone '{zone.Name}' set to '{dimming_type}'")
            return float(dimming_type)  # Return first element
        else:
            raise ValueError(
                "Could not resolve more than one illuminance targets for Zone {}. "
                "Make sure there is only one".format(zone.Name)
            )
    else:
        # Else, there are no dimming controls => set to "Off".
        log(
            "No illuminance target found for zone {}. Setting to default 500 "
            "lux".format(zone.Name),
            lg.DEBUG,
        )
        return 500



################################################################################
# Module: __init__.py
# Description: Archetypal: Retrieve, construct and analyse building archetypes
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

# Version of the package
__version__ = "1.3.1"

# warn if a newer version of archetypal is available
from outdated import warn_if_outdated
from .utils import warn_if_not_compatible

warn_if_outdated("archetypal", __version__)
warn_if_not_compatible()

from .utils import *
from .simple_glazing import *
from .energyseries import EnergySeries
from .energydataframe import EnergyDataFrame
from .reportdata import ReportData
from .tabulardata import TabularData
from .idfclass import *
from .schedule import Schedule
from .dataportal import *
from .plot import *
from .trnsys import *
from .template import *
from .umi_template import *
from .cli import *



################################################################################
# Module: template.__init__.py
# Description: Template module for everything UMI template.
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

from .umi_base import *
from .gas_material import *
from .glazing_material import *
from .opaque_material import *
from .opaque_construction import *
from .schedule import *
from .window import *
from .structure import *
from .conditioning import *
from .load import *
from .ventilation import *
from .dhw import *
from .zone import *
from .building_template import *



################################################################################
# Module: archetypal.template
# Description:
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import collections

from archetypal import log
from archetypal.template import UmiBase, Unique


class OpaqueMaterial(UmiBase, metaclass=Unique):
    """Use this component to create a custom opaque material.

    .. image:: ../images/template/materials-opaque.png
    """

    def __init__(
        self,
        Conductivity,
        SpecificHeat,
        SolarAbsorptance=0.7,
        ThermalEmittance=0.9,
        VisibleAbsorptance=0.7,
        Roughness="Rough",
        Cost=0,
        Density=1,
        MoistureDiffusionResistance=50,
        EmbodiedCarbon=0.45,
        EmbodiedEnergy=0,
        TransportCarbon=0,
        TransportDistance=0,
        TransportEnergy=0,
        SubstitutionRatePattern=None,
        SubstitutionTimestep=20,
        **kwargs
    ):
        """A custom opaque material.

        Args:
            Conductivity (float): A number representing the conductivity of the
                material in W/m-K. This is essentially the heat flow in Watts
                across one meter thick of the material when the temperature
                difference on either side is 1 Kelvin. Modeling layers with
                conductivity higher than 5.0 W/(m-K) is not recommended.
            SpecificHeat (float): A number representing the specific heat
                capacity of the material in J/kg-K. This is essentially the
                number of joules needed to raise one kg of the material by 1
                degree Kelvin. Only values of specific heat of 100 or larger are
                allowed. Typical ranges are from 800 to 2000 J/(kg-K).
            SolarAbsorptance (float): An number between 0 and 1 that represents
                the absorptance of solar radiation by the material. The default
                is set to 0.7, which is common for most non-metallic materials.
            ThermalEmittance (float): An number between 0 and 1 that represents
                the thermal abstorptance of the material. The default is set to
                0.9, which is common for most non-metallic materials. For long
                wavelength radiant exchange, thermal emissivity and thermal
                emittance are equal to thermal absorptance.
            VisibleAbsorptance (float): An number between 0 and 1 that
                represents the absorptance of visible light by the material.
                The default is set to 0.7, which is common for most non-metallic
                materials.
            Roughness (str): A text value that indicated the roughness of your
                material. This can be either "VeryRough", "Rough",
                "MediumRough", "MediumSmooth", "Smooth", and "VerySmooth". The
                default is set to "Rough".
            Cost: # todo: define parameter
            Density (float): A number representing the density of the material
                in kg/m3. This is essentially the mass of one cubic meter of the
                material.
            MoistureDiffusionResistance: # todo: defined parameter
            EmbodiedCarbon: # todo: define parameter
            EmbodiedEnergy: # todo: define parameter
            TransportCarbon: # todo: define parameter
            TransportDistance: # todo: define parameter
            TransportEnergy: # todo: define parameter
            SubstitutionRatePattern: # todo: define parameter
            SubstitutionTimestep: # todo: define parameter
            **kwargs:
        """
        super(OpaqueMaterial, self).__init__(**kwargs)

        if SubstitutionRatePattern is None:
            SubstitutionRatePattern = [0.5, 1]
        self.Conductivity = Conductivity
        self.Roughness = Roughness
        self.SolarAbsorptance = SolarAbsorptance
        self.SpecificHeat = SpecificHeat
        self.ThermalEmittance = ThermalEmittance
        self.VisibleAbsorptance = VisibleAbsorptance
        self.TransportCarbon = TransportCarbon
        self.TransportDistance = TransportDistance
        self.TransportEnergy = TransportEnergy
        self.SubstitutionRatePattern = SubstitutionRatePattern
        self.SubstitutionTimestep = SubstitutionTimestep
        self.Cost = Cost
        self.Density = Density
        self.EmbodiedCarbon = EmbodiedCarbon
        self.EmbodiedEnergy = EmbodiedEnergy
        self.MoistureDiffusionResistance = MoistureDiffusionResistance

        self._thickness = kwargs.get("Thickness", None)

    def __add__(self, other):
        """Overload + to implement self.combine.

        Args:
            other (OpaqueMaterial):
        """
        return self.combine(other)

    def __hash__(self):
        return hash((self.__class__.__name__, self.Name))

    def __eq__(self, other):
        if not isinstance(other, OpaqueMaterial):
            return False
        else:
            return all(
                [
                    self.Conductivity == other.Conductivity,
                    self.SpecificHeat == other.SpecificHeat,
                    self.SolarAbsorptance == other.SolarAbsorptance,
                    self.ThermalEmittance == other.ThermalEmittance,
                    self.VisibleAbsorptance == other.VisibleAbsorptance,
                    self.Roughness == other.Roughness,
                    self.Cost == other.Cost,
                    self.Density == other.Density,
                    self.MoistureDiffusionResistance
                    == self.MoistureDiffusionResistance,
                    self.EmbodiedCarbon == other.EmbodiedCarbon,
                    self.EmbodiedEnergy == other.EmbodiedEnergy,
                    self.TransportCarbon == other.TransportCarbon,
                    self.TransportDistance == other.TransportDistance,
                    self.TransportEnergy == other.TransportEnergy,
                    self.SubstitutionRatePattern == other.SubstitutionRatePattern,
                    self.SubstitutionTimestep == other.SubstitutionTimestep,
                ]
            )

    @classmethod
    def generic(cls):
        """generic plaster board"""
        return cls(
            Conductivity=1.39,
            SpecificHeat=1085,
            Density=2000,
            Name="generic_plaster_board",
        )

    def combine(self, other, weights=None):
        """Combine two OpaqueMaterial objects.

        Args:
            weights (list-like, optional): A list-like object of len 2. If None,
                the density of the OpaqueMaterial of each objects is used as
                a weighting factor.
            other (OpaqueMaterial): The other OpaqueMaterial object the
                combine with.

        Returns:
            OpaqueMaterial: A new combined object made of self + other.
        """
        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        if not weights:
            log(
                'using OpaqueMaterial density as weighting factor in "{}" '
                "combine.".format(self.__class__.__name__)
            )
            weights = [self.Density, other.Density]

        meta = self._get_predecessors_meta(other)
        new_obj = OpaqueMaterial(
            **meta,
            Conductivity=self._float_mean(other, "Conductivity", weights),
            Roughness=self._str_mean(other, attr="Roughness", append=False),
            SolarAbsorptance=self._float_mean(other, "SolarAbsorptance", weights),
            SpecificHeat=self._float_mean(other, "SpecificHeat"),
            ThermalEmittance=self._float_mean(other, "ThermalEmittance", weights),
            VisibleAbsorptance=self._float_mean(other, "VisibleAbsorptance", weights),
            TransportCarbon=self._float_mean(other, "TransportCarbon", weights),
            TransportDistance=self._float_mean(other, "TransportDistance", weights),
            TransportEnergy=self._float_mean(other, "TransportEnergy", weights),
            SubstitutionRatePattern=self._float_mean(
                other, "SubstitutionRatePattern", weights=None
            ),
            SubstitutionTimestep=self._float_mean(
                other, "SubstitutionTimestep", weights
            ),
            Cost=self._float_mean(other, "Cost", weights),
            Density=self._float_mean(other, "Density", weights),
            EmbodiedCarbon=self._float_mean(other, "EmbodiedCarbon", weights),
            EmbodiedEnergy=self._float_mean(other, "EmbodiedEnergy", weights),
            MoistureDiffusionResistance=self._float_mean(
                other, "MoistureDiffusionResistance", weights
            )
        )
        new_obj._predecessors.extend(self.predecessors + other.predecessors)
        return new_obj

    def to_json(self):
        """Convert class properties to dict"""
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["MoistureDiffusionResistance"] = self.MoistureDiffusionResistance
        data_dict["Roughness"] = self.Roughness
        data_dict["SolarAbsorptance"] = self.SolarAbsorptance
        data_dict["SpecificHeat"] = self.SpecificHeat
        data_dict["ThermalEmittance"] = self.ThermalEmittance
        data_dict["VisibleAbsorptance"] = self.VisibleAbsorptance
        data_dict["Conductivity"] = self.Conductivity
        data_dict["Cost"] = self.Cost
        data_dict["Density"] = self.Density
        data_dict["EmbodiedCarbon"] = self.EmbodiedCarbon
        data_dict["EmbodiedEnergy"] = self.EmbodiedEnergy
        data_dict["SubstitutionRatePattern"] = self.SubstitutionRatePattern
        data_dict["SubstitutionTimestep"] = self.SubstitutionTimestep
        data_dict["TransportCarbon"] = self.TransportCarbon
        data_dict["TransportDistance"] = self.TransportDistance
        data_dict["TransportEnergy"] = self.TransportEnergy
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = self.Name

        return data_dict

    @classmethod
    def from_epbunch(cls, epbunch, **kwargs):
        """Create an OpaqueMaterial from an IDF "Material", "Material:NoMAss",
        or "Material:AirGap" element.

        Hint:
            (From EnergyPlus Manual): When a user enters such a “no mass”
            material into EnergyPlus, internally the properties of this layer
            are converted to approximate the properties of air (density,
            specific heat, and conductivity) with the thickness adjusted to
            maintain the user’s desired R-Value. This allowed such layers to be
            handled internally in the same way as other layers without any
            additional changes to the code. This solution was deemed accurate
            enough as air has very little thermal mass and it made the coding of
            the state space method simpler.

            For Material:AirGap, a similar strategy is used, with the
            exception that solar properties (solar and visible absorptance and
            emittance) are assumed null.

        Args:
            epbunch (EpBunch): EP-Construction object
            **kwargs:
        """
        if epbunch.key.upper() == "MATERIAL":
            # do MATERIAL
            Name = epbunch.Name
            Conductivity = epbunch.Conductivity
            Density = epbunch.Density
            Roughness = epbunch.Roughness
            SolarAbsorptance = epbunch.Solar_Absorptance
            SpecificHeat = epbunch.Specific_Heat
            ThermalEmittance = epbunch.Thermal_Absorptance
            VisibleAbsorptance = epbunch.Visible_Absorptance
            Thickness = epbunch.Thickness
            return cls(
                Conductivity=Conductivity,
                Density=Density,
                Roughness=Roughness,
                SolarAbsorptance=SolarAbsorptance,
                SpecificHeat=SpecificHeat,
                ThermalEmittance=ThermalEmittance,
                VisibleAbsorptance=VisibleAbsorptance,
                Thickness=Thickness,
                Name=Name,
                idf=epbunch.theidf,
                **kwargs
            )
        elif epbunch.key.upper() == "MATERIAL:NOMASS":
            # do MATERIAL:NOMASS. Assume properties of air.
            Name = epbunch.Name
            Conductivity = 0.02436  # W/mK, dry air at 0 °C and 100 kPa.
            Density = 1.2754  # dry air at 0 °C and 100 kPa.
            SpecificHeat = 100.5  # J/kg-K, dry air at 0 °C and 100 kPa.
            Thickness = Conductivity * epbunch.Thermal_Resistance
            Roughness = epbunch.Roughness
            SolarAbsorptance = epbunch.Solar_Absorptance
            ThermalEmittance = epbunch.Thermal_Absorptance
            VisibleAbsorptance = epbunch.Visible_Absorptance
            return cls(
                Conductivity=Conductivity,
                Density=Density,
                Roughness=Roughness,
                SolarAbsorptance=SolarAbsorptance,
                SpecificHeat=SpecificHeat,
                ThermalEmittance=ThermalEmittance,
                VisibleAbsorptance=VisibleAbsorptance,
                Thickness=Thickness,
                Name=Name,
                idf=epbunch.theidf,
                **kwargs
            )
        elif epbunch.key.upper() == "MATERIAL:AIRGAP":

            Name = epbunch.Name
            Conductivity = 0.02436  # W/mK, dry air at 0 °C and 100 kPa.
            Density = 1.2754  # dry air at 0 °C and 100 kPa.
            SpecificHeat = 100.5  # J/kg-K, dry air at 0 °C and 100 kPa.
            Thickness = Conductivity * epbunch.Thermal_Resistance
            Roughness = "Smooth"
            return cls(
                Conductivity=Conductivity,
                Roughness=Roughness,
                SpecificHeat=SpecificHeat,
                Thickness=Thickness,
                Density=Density,
                Name=Name,
                SolarAbsorptance=0,
                ThermalEmittance=0,
                VisibleAbsorptance=0,
                idf=epbunch.theidf,
                **kwargs
            )
        else:
            raise NotImplementedError(
                "Material '{}' of type '{}' is not yet "
                "supported. Please contact package "
                "authors".format(epbunch.Name, epbunch.key)
            )



import copy
import logging as lg
import os
import time
import warnings
from datetime import timedelta

import numpy as np
import pandas as pd
import tsam.timeseriesaggregation as tsam
from matplotlib import pyplot as plt, cm
from matplotlib.colors import LightSource
from pandas import Series, DataFrame, concat, MultiIndex, date_range
from sklearn import preprocessing

import archetypal
from archetypal import log, rmse, piecewise, settings


class EnergySeries(Series):
    """A Series object designed to store energy related data."""

    @property
    def _constructor(self):
        return EnergySeries

    _metadata = [
        "name",
        "bin_edges_",
        "bin_scaling_factors_",
        "profile_type",
        "base_year",
        "frequency",
        "units",
        "sort_values",
        "to_units",
        "converted_",
        "concurrent_sort_",
    ]

    def __new__(
        cls,
        data,
        frequency=None,
        units=None,
        profile_type="undefinded",
        index=None,
        dtype=None,
        copy=True,
        name=None,
        fastpath=False,
        base_year=2018,
        normalize=False,
        sort_values=False,
        ascending=False,
        archetypes=None,
        concurrent_sort=False,
        to_units=None,
        use_timeindex=False,
    ):
        """
        Args:
            data:
            frequency:
            units:
            profile_type:
            index:
            dtype:
            copy:
            name:
            fastpath:
            base_year:
            normalize:
            sort_values:
            ascending:
            archetypes:
            concurrent_sort:
            to_units:
            use_timeindex:
        """
        self = super(EnergySeries, cls).__new__(cls)
        return self

    def __init__(
        self,
        data,
        frequency=None,
        units=None,
        profile_type="undefinded",
        index=None,
        dtype=None,
        copy=True,
        name=None,
        fastpath=False,
        base_year=2018,
        normalize=False,
        sort_values=False,
        ascending=False,
        archetypes=None,
        concurrent_sort=False,
        to_units=None,
        use_timeindex=False,
    ):
        """
        Args:
            data:
            frequency:
            units:
            profile_type:
            index:
            dtype:
            copy:
            name:
            fastpath:
            base_year:
            normalize:
            sort_values:
            ascending:
            archetypes:
            concurrent_sort:
            to_units:
            use_timeindex:
        """
        super(EnergySeries, self).__init__(
            data=data, index=index, dtype=dtype, name=name, copy=copy, fastpath=fastpath
        )
        self.bin_edges_ = None
        self.bin_scaling_factors_ = None
        self.profile_type = profile_type
        self.frequency = frequency
        self.base_year = base_year
        self.units = settings.unit_registry.parse_expression(units).units
        self.archetypes = archetypes
        self.to_units = settings.unit_registry.parse_expression(to_units).units
        self.converted_ = False
        self.concurrent_sort_ = concurrent_sort
        # handle sorting of the data
        if sort_values:
            self.is_sorted = True
            if concurrent_sort:
                self.concurrent_sort(ascending=ascending, inplace=True)
            else:
                self.sort_values(ascending=ascending, inplace=True)
                self.reset_index(drop=True, inplace=True)
        else:
            self.is_sorted = False

        # handle archetype names
        if isinstance(self.index, MultiIndex):
            self.archetypes = list(set(self.index.get_level_values(level=0)))
        else:
            self.archetypes = None

        # handle normalization
        if normalize:
            self.normalize(inplace=True)

        # handle unit conversion
        if to_units and units:
            self.unit_conversion(to_units=to_units, inplace=True)

        # handle DateTimeIndex
        if index is None and use_timeindex:
            start_date = str(self.base_year) + "0101"
            if isinstance(self.index, MultiIndex):
                newindex = self.index  # todo: finish this
            else:
                newindex = pd.date_range(
                    start=start_date, freq=self.frequency, periods=len(self)
                )
            self.index = newindex

    @classmethod
    def from_sqlite(
        cls,
        df,
        name=None,
        base_year=2018,
        units=None,
        normalize=False,
        sort_values=False,
        ascending=False,
        concurrent_sort=False,
        to_units=None,
        agg_func="sum",
    ):
        """Create a.

        Args:
            df (DataFrame):
            name:
            base_year:
            units:
            normalize:
            sort_values:
            ascending:
            concurrent_sort:
            to_units:
            agg_func (callable): The aggregation function to use in the case
                that multiple values have the same index value. If a function,
                must either work when passed a DataFrame or when passed to
                DataFrame.apply. For a DataFrame, can pass a dict, if the keys
                are DataFrame column names.

                Accepted Combinations are:
                    - string function name
                    - function
                    - list of functions
                    - dict of column names -> functions (or list of functions)
        """
        index = pd.to_datetime(
            {
                "year": base_year,
                "month": df.Month,
                "day": df.Day,
                "hour": df.Hour,
                "minute": df.Minute,
            }
        )
        # Adjust timeindex by timedelta
        index -= df.Interval.apply(lambda x: timedelta(minutes=x))
        index = pd.DatetimeIndex(index)
        # get data
        data = df.Value
        data.index = index
        units = [units] if units else set(df.Units)
        if len(units) > 1:
            raise ValueError("The DataFrame contains mixed units: {}".format(units))
        else:
            units = next(iter(units), None)
        # group data by index value (level=0) using the agg_func
        if agg_func:
            grouped_Data = data.groupby(level=0).agg(agg_func)
        else:
            df["DateTimeIndex"] = index
            grouped_Data = df.set_index(["DateTimeIndex", "Name"]).Value
        # Since we create the index, use_timeindex must be false
        return cls(
            grouped_Data.values,
            name=name,
            units=units,
            index=grouped_Data.index,
            use_timeindex=False,
            base_year=base_year,
            normalize=normalize,
            sort_values=sort_values,
            ascending=ascending,
            concurrent_sort=concurrent_sort,
            to_units=to_units,
        )

    def unit_conversion(self, to_units=None, inplace=False):
        """returns the multiplier to convert units

        Args:
            to_units (pint.Unit):
            inplace:
        """
        reg = settings.unit_registry
        if to_units is None:
            to_units = self.to_units
        else:
            to_units = reg.parse_expression(to_units).units
        cdata = reg.Quantity(self.values, self.units).to(to_units).m
        result = self.apply(lambda x: x)
        result.update(pd.Series(cdata, index=result.index))
        result.__class__ = EnergySeries
        result.converted_ = True
        result.units = to_units
        if inplace:
            self._update_inplace(result)
            self.__finalize__(result)
        else:
            return result

    def concurrent_sort(self, ascending=False, inplace=False, level=0):
        """
        Args:
            ascending:
            inplace:
            level:
        """
        if isinstance(self.index, MultiIndex):
            concurrent = self.unstack(level=level)
            concurrent_sum = concurrent.sum(axis=1)

            sortedIdx = concurrent_sum.sort_values(ascending=ascending).index

            result = concurrent.loc[sortedIdx, :]
            result.index = concurrent.index
            result = result.stack().swaplevel()

            if inplace:
                self.update(result)
                self.__finalize__(result)
            else:
                return result  # todo: make sure results has all the metadata

    def normalize(self, feature_range=(0, 1), inplace=False):
        """Returns a normalized EnergySeries

        Args:
            feature_range:
            inplace:
        """
        scaler = preprocessing.MinMaxScaler(feature_range=feature_range)
        if self.archetypes:
            result = concat(
                {
                    name: Series(
                        scaler.fit_transform(sub.values.reshape(-1, 1)).ravel()
                    )
                    for name, sub in self.groupby(level=0)
                }
            ).sort_index()
            result = self._constructor(result)
        else:
            result = Series(scaler.fit_transform(self.values.reshape(-1, 1)).ravel())
            result = self._constructor(result)
            result.units = settings.unit_registry.dimensionless
        if inplace:
            self._update_inplace(result)
        else:
            return result  # todo: make sure result has all the metadata

    def ldc_source(self, SCOPH=4, SCOPC=4):
        """Returns the Load Duration Curve from the source side of theoretical
        Heat Pumps

        Args:
            SCOPH: Seasonal COP in Heating
            SCOPC: Seasonal COP in Cooling

        Returns:
            (EnergySeries) Load Duration Curve
        """

        result = self.ldc.apply(
            lambda x: x * (1 - 1 / SCOPH) if x > 0 else x * (1 + 1 / SCOPC)
        )
        return result

    def source_side(self, SCOPH=None, SCOPC=None):
        """Returns the Source Side EnergySeries given a Seasonal COP. Negative
        values are considered like Cooling Demand.

        Args:
            SCOPH: Seasonal COP in Heating
            SCOPC: Seasonal COP in Cooling

        Returns:
            (EnergySeries) Load Duration Curve
        """
        if SCOPC or SCOPH:
            result = self.apply(
                lambda x: x * (1 - 1 / SCOPH) if SCOPH else x * (1 + 1 / SCOPC)
            )
            return result
        else:
            raise ValueError("Please provide a SCOPH or a SCOPC")

    def discretize_tsam(
        self,
        resolution=None,
        noTypicalPeriods=10,
        hoursPerPeriod=24,
        clusterMethod="hierarchical",
        evalSumPeriods=False,
        sortValues=False,
        sameMean=False,
        rescaleClusterPeriods=True,
        weightDict=None,
        extremePeriodMethod="None",
        solver="glpk",
        roundOutput=None,
        addPeakMin=None,
        addPeakMax=None,
        addMeanMin=None,
        addMeanMax=None,
    ):
        """uses tsam

        Args:
            resolution:
            noTypicalPeriods:
            hoursPerPeriod:
            clusterMethod:
            evalSumPeriods:
            sortValues:
            sameMean:
            rescaleClusterPeriods:
            weightDict:
            extremePeriodMethod:
            solver:
            roundOutput:
            addPeakMin:
            addPeakMax:
            addMeanMin:
            addMeanMax:
        """
        try:
            import tsam.timeseriesaggregation as tsam
        except ImportError:
            raise ImportError("tsam is required for discretize_tsam()")
        if not isinstance(self.index, pd.DatetimeIndex):
            raise TypeError("To use tsam, index of series must be a " "DateTimeIndex")
        if isinstance(self, Series):
            timeSeries = pd.DataFrame(self)
        else:
            timeSeries = self.copy()
        agg = tsam.TimeSeriesAggregation(
            timeSeries,
            resolution=resolution,
            noTypicalPeriods=noTypicalPeriods,
            hoursPerPeriod=hoursPerPeriod,
            clusterMethod=clusterMethod,
            evalSumPeriods=evalSumPeriods,
            sortValues=sortValues,
            sameMean=sameMean,
            rescaleClusterPeriods=rescaleClusterPeriods,
            weightDict=weightDict,
            extremePeriodMethod=extremePeriodMethod,
            solver=solver,
            roundOutput=roundOutput,
            addPeakMin=addPeakMin,
            addPeakMax=addPeakMax,
            addMeanMin=addMeanMin,
            addMeanMax=addMeanMax,
        )

        agg.createTypicalPeriods()
        results = agg.predictOriginalData()
        results = EnergySeries(results.iloc[:, 0])
        # results._internal_names += agg.clusterOrder
        return results.__finalize__(self)

    def discretize(self, n_bins=3, inplace=False):
        """Retruns a discretized pandas.Series

        Args:
            n_bins (int): Number of bins or steps to discretize the function
            inplace (bool): if True, perform operation in-place
        """
        try:
            from scipy.optimize import minimize
            from itertools import chain
        except ImportError:
            raise ImportError(
                "The sklearn package must be installed to " "use this optional feature."
            )
        if self.archetypes:
            # if multiindex, group and apply operation on each group.
            # combine at the end
            results = {}
            edges = {}
            ampls = {}
            for name, sub in self.groupby(level=0):
                hour_of_min = sub.time_at_min[1]

                sf = [1 / (i * 1.01) for i in range(1, n_bins + 1)]
                sf.extend([sub.min()])
                sf_bounds = [(0, sub.max()) for i in range(0, n_bins + 1)]
                hours = [
                    hour_of_min - hour_of_min * 1 / (i * 1.01)
                    for i in range(1, n_bins + 1)
                ]
                # Todo hours need to work fow datetime index
                hours.extend([len(sub)])
                hours_bounds = [(0, len(sub)) for i in range(0, n_bins + 1)]

                start_time = time.time()
                log("discretizing EnergySeries {}".format(name), lg.DEBUG)
                res = minimize(
                    rmse,
                    np.array(hours + sf),
                    args=(sub.values),
                    method="L-BFGS-B",
                    bounds=hours_bounds + sf_bounds,
                    options=dict(disp=True),
                )
                log(
                    "Completed discretization in {:,.2f} seconds".format(
                        time.time() - start_time
                    ),
                    lg.DEBUG,
                )
                edges[name] = res.x[0 : n_bins + 1]
                ampls[name] = res.x[n_bins + 1 :]
                results[name] = Series(piecewise(res.x))
            self.bin_edges_ = Series(edges).apply(Series)
            self.bin_scaling_factors_ = DataFrame(ampls)

            result = concat(results)
        else:
            hour_of_min = self.time_at_min

            sf = [1 / (i * 1.01) for i in range(1, n_bins + 1)]
            sf.extend([self.min()])
            sf_bounds = [(0, self.max()) for i in range(0, n_bins + 1)]
            hours = [
                hour_of_min - hour_of_min * 1 / (i * 1.01) for i in range(1, n_bins + 1)
            ]
            hours.extend([len(self)])
            hours_bounds = [(0, len(self)) for i in range(0, n_bins + 1)]

            start_time = time.time()
            # log('discretizing EnergySeries {}'.format(name), lg.DEBUG)
            res = minimize(
                rmse,
                np.array(hours + sf),
                args=(self.values),
                method="L-BFGS-B",
                bounds=hours_bounds + sf_bounds,
                options=dict(disp=True),
            )
            log(
                "Completed discretization in {:,.2f} seconds".format(
                    time.time() - start_time
                ),
                lg.DEBUG,
            )
            edges = res.x[0 : n_bins + 1]
            ampls = res.x[n_bins + 1 :]
            result = Series(piecewise(res.x))
            bin_edges = Series(edges).apply(Series)
            self.bin_edges_ = bin_edges
            bin_edges.loc[-1, 0] = 0
            bin_edges.sort_index(inplace=True)
            bin_edges = bin_edges.diff().dropna()
            bin_edges = bin_edges.round()
            self.bin_scaling_factors_ = DataFrame(
                {"duration": bin_edges[0], "scaling factor": ampls}
            )
            self.bin_scaling_factors_.index = np.round(edges).astype(int)

        if inplace:
            self.update(result)
            self.__class__ = EnergySeries
            self.__finalize__(result)
        else:
            result.__class__ = EnergySeries
            return result.__finalize__(self)

    def unstack(self, level=-1, fill_value=None):
        """
        Args:
            level:
            fill_value:
        """
        from pandas.core.reshape.reshape import unstack

        result = unstack(self, level, fill_value)
        result.__class__ = archetypal.EnergyDataFrame
        return result.__finalize__(self)

    def stack(self, level=-1, dropna=True):
        """
        Args:
            level:
            dropna:
        """
        from pandas.core.reshape.reshape import stack, stack_multiple

        if isinstance(level, (tuple, list)):
            result = stack_multiple(self, level, dropna=dropna)
            return self.__finalize__(result)
        else:
            result = stack(self, level, dropna=dropna)
            return self.__finalize__(result)

    def plot3d(self, *args, **kwargs):
        """Generate a plot of the EnergySeries.

        Wraps the ``plot_energyseries()`` function, and documentation is
        copied from there.

        Args:
            *args:
            **kwargs:
        """
        return plot_energyseries(self, *args, **kwargs)

    def plot2d(self, *args, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        return plot_energyseries_map(self, **kwargs)

    # @property
    # def units(self):
    #     return self.units.units

    @property
    def p_max(self):
        if isinstance(self.index, MultiIndex):
            return self.groupby(level=0).max()
        else:
            return self.max()

    @property
    def monthly(self):
        if isinstance(self.index, MultiIndex):
            return self.groupby(level=0).max()
        else:
            datetimeindex = date_range(
                freq=self.frequency,
                start="{}-01-01".format(self.base_year),
                periods=self.size,
            )
            self_copy = self.copy()
            self_copy.index = datetimeindex
            self_copy = self_copy.resample("M").mean()
            self_copy.frequency = "M"
            return EnergySeries(self_copy, frequency="M", units=self.units)

    @property
    def capacity_factor(self):
        max = self.max()
        mean = self.mean()
        return mean / max

    @property
    def bin_edges(self):
        return self.bin_edges_

    @property
    def time_at_min(self):
        return self.idxmin()

    @property
    def bin_scaling_factors(self):
        return self.bin_scaling_factors_

    @property
    def duration_scaling_factor(self):
        return list(map(tuple, self.bin_scaling_factors.values))

    @property
    def ldc(self):
        nb_points = len(self)
        newdata = self.sort_values(ascending=False).reset_index(drop=True)
        return newdata.__finalize__(self)

    @property
    def nseries(self):
        if self.data.ndim == 1:
            return 1
        else:
            return self.data.shape[1]


def save_and_show(
    fig, ax, save, show, close, filename, file_format, dpi, axis_off, extent
):
    """Save a figure to disk and show it, as specified.

    Args:
        fig (matplotlib.figure.Figure): the figure
        ax (matplotlib.axes.Axes or list(matplotlib.axes.Axes)): the axes
        save (bool): whether to save the figure to disk or not
        show (bool): whether to display the figure or not
        close (bool): close the figure (only if show equals False) to prevent
            display
        filename (string): the name of the file to save
        file_format (string): the format of the file to save (e.g., 'jpg',
            'png', 'svg')
        dpi (int): the resolution of the image file if saving (Dots per inch)
        axis_off (bool): if True matplotlib axis was turned off by plot_graph so
            constrain the saved figure's extent to the interior of the axis
        extent:

    Returns:
        (tuple) fig, ax
    """
    # save the figure if specified

    if save:
        start_time = time.time()

        # create the save folder if it doesn't already exist
        if not os.path.exists(settings.imgs_folder):
            os.makedirs(settings.imgs_folder)
        path_filename = os.path.join(
            settings.imgs_folder, os.extsep.join([filename, file_format])
        )

        if not isinstance(ax, (np.ndarray, list)):
            ax = [ax]
        if file_format == "svg":
            for ax in ax:
                # if the file_format is svg, prep the fig/ax a bit for saving
                ax.axis("off")
                ax.set_position([0, 0, 1, 1])
                ax.patch.set_alpha(0.0)
            fig.patch.set_alpha(0.0)
            fig.savefig(
                path_filename,
                bbox_inches=0,
                format=file_format,
                facecolor=fig.get_facecolor(),
                transparent=True,
            )
        else:
            if extent is None:
                if len(ax) == 1:
                    if axis_off:
                        for ax in ax:
                            # if axis is turned off, constrain the saved
                            # figure's extent to the interior of the axis
                            extent = ax.get_window_extent().transformed(
                                fig.dpi_scale_trans.inverted()
                            )
                else:
                    extent = "tight"
            fig.savefig(
                path_filename,
                dpi=dpi,
                bbox_inches=extent,
                format=file_format,
                facecolor=fig.get_facecolor(),
                transparent=True,
            )
        log(
            "Saved the figure to disk in {:,.2f} seconds".format(
                time.time() - start_time
            )
        )

    # show the figure if specified
    if show:
        start_time = time.time()
        plt.show()
        # fig.show()
        log("Showed the plot in {:,.2f} seconds".format(time.time() - start_time))
    # if show=False, close the figure if close=True to prevent display
    elif close:
        plt.close()

    return fig, ax


def plot_energyseries(
    energy_series,
    kind="polygon",
    axis_off=True,
    cmap=None,
    fig_height=None,
    fig_width=6,
    show=True,
    view_angle=-60,
    save=False,
    close=False,
    dpi=300,
    file_format="png",
    color=None,
    axes=None,
    vmin=None,
    vmax=None,
    filename=None,
    timeStepsPerPeriod=24,
    **kwargs
):
    """
    Args:
        energy_series (EnergySeries):
        kind (str):
        axis_off (bool):
        cmap:
        fig_height (float):
        fig_width (float):
        show (bool):
        view_angle (float):
        save (bool):
        close (bool):
        dpi (int):
        file_format (str):
        color (str):
        axes:
        vmin (float):
        vmax (float):
        filename (str):
        timeStepsPerPeriod (int): The number of discrete timesteps which
            describe one period.
        **kwargs:
    """
    if energy_series.empty:
        warnings.warn(
            "The EnergyProgile you are attempting to plot is "
            "empty. Nothing has been displayed.",
            UserWarning,
        )
        return axes

    import matplotlib.pyplot as plt

    # noinspection PyUnresolvedReferences
    from mpl_toolkits.mplot3d import Axes3D

    if isinstance(energy_series.index, pd.MultiIndex):
        groups = energy_series.groupby(level=0)
        nax = len(groups)
    else:
        nax = 1
        groups = [("unnamed", energy_series)]

    if fig_height is None:
        fig_height = fig_width * nax

    # Set up plot
    fig, axes = plt.subplots(
        nax,
        1,
        subplot_kw=dict(projection="3d"),
        figsize=(fig_width, fig_height),
        dpi=dpi,
    )
    if not isinstance(axes, np.ndarray):
        axes = [axes]

    for ax, (name, profile) in zip(axes, groups):
        values = profile.values

        vmin = values.min() if vmin is None else vmin
        vmax = values.max() if vmax is None else vmax

        if kind == "polygon":
            import tsam.timeseriesaggregation as tsam

            z, _ = tsam.unstackToPeriods(profile, timeStepsPerPeriod=timeStepsPerPeriod)
            nrows, ncols = z.shape

            xs = z.columns
            zs = z.index.values

            verts = []
            for i in zs:
                ys = z.iloc[int(i), :]
                verts.append(_polygon_under_graph(xs, ys))

            _plot_poly_collection(
                ax,
                verts,
                zs,
                edgecolors=kwargs.get("edgecolors", None),
                facecolors=kwargs.get("facecolors", None),
                linewidths=kwargs.get("linewidths", None),
                cmap=cmap,
            )
        elif kind == "surface":
            import tsam.timeseriesaggregation as tsam

            z, _ = tsam.unstackToPeriods(profile, timeStepsPerPeriod=timeStepsPerPeriod)
            nrows, ncols = z.shape
            x = z.columns
            y = z.index.values

            x, y = np.meshgrid(x, y)
            _plot_surface(ax, x, y, z.values, cmap=cmap, **kwargs)
        elif kind == "contour":
            import tsam.timeseriesaggregation as tsam

            z, _ = tsam.unstackToPeriods(profile, timeStepsPerPeriod=timeStepsPerPeriod)
            nrows, ncols = z.shape
            x = z.columns
            y = z.index.values

            x, y = np.meshgrid(x, y)
            _plot_contour(ax, x, y, z.values, cmap=cmap, **kwargs)
        else:
            raise NameError('plot kind "{}" is not supported'.format(kind))

        if filename is None:
            filename = "unnamed"

        # set the extent of the figure
        ax.set_xlim3d(-1, ncols)
        ax.set_xlabel(kwargs.get("xlabel", "hour of day"))
        ax.set_ylim3d(-1, nrows)
        ax.set_ylabel(kwargs.get("ylabel", "day of year"))
        ax.set_zlim3d(vmin, vmax)
        z_label = "{} [{:~P}]".format(
            energy_series.name if energy_series.name is not None else "Z",
            energy_series.units,
        )
        ax.set_zlabel(z_label)

        # configure axis appearance
        xaxis = ax.xaxis
        yaxis = ax.yaxis
        zaxis = ax.zaxis

        xaxis.get_major_formatter().set_useOffset(False)
        yaxis.get_major_formatter().set_useOffset(False)
        zaxis.get_major_formatter().set_useOffset(False)

        # if axis_off, turn off the axis display set the margins to zero and
        # point the ticks in so there's no space around the plot
        if axis_off:
            ax.axis("off")
            ax.margins(0)
            ax.tick_params(which="both", direction="in")
            xaxis.set_visible(False)
            yaxis.set_visible(False)
            zaxis.set_visible(False)
            fig.canvas.draw()
        if view_angle is not None:
            ax.view_init(30, view_angle)
            ax.set_proj_type(kwargs.get("proj_type", "persp"))
            fig.canvas.draw()
    fig, axes = save_and_show(
        fig=fig,
        ax=axes,
        save=save,
        show=show,
        close=close,
        filename=filename,
        file_format=file_format,
        dpi=dpi,
        axis_off=axis_off,
        extent=None,
    )
    return fig, axes


def plot_energyseries_map(
    data,
    periodlength=24,
    subplots=False,
    vmin=None,
    vmax=None,
    axis_off=True,
    cmap="RdBu",
    fig_height=None,
    fig_width=6,
    show=True,
    view_angle=-60,
    save=False,
    close=False,
    dpi=300,
    file_format="png",
    color=None,
    ax=None,
    filename="untitled",
    extent="tight",
    sharex=False,
    sharey=False,
    layout=None,
    layout_type="vertical",
    **kwargs
):
    """
    Args:
        data (EnergySeries or EnergyDataFrame):
        periodlength:
        subplots:
        vmin:
        vmax:
        axis_off:
        cmap:
        fig_height:
        fig_width:
        show:
        view_angle:
        save:
        close:
        dpi:
        file_format:
        color:
        ax:
        filename:
        extent:
        sharex:
        sharey:
        layout:
        layout_type:
        **kwargs:
    """
    if fig_height is None:
        fig_height = fig_width / 3
    figsize = (fig_width, fig_height)

    if not ax:
        if subplots:
            if isinstance(data, EnergySeries):
                data = data.unstack(level=0)
            n = data.shape[1]
        else:
            n = 1
        fig, axes = plt.subplots(
            nrows=n, ncols=1, figsize=(fig_width, fig_height), dpi=dpi
        )
    else:
        fig = ax.get_figure()
        if figsize is not None:
            fig.set_size_inches(figsize)
        axes = ax

    stacked, timeindex = tsam.unstackToPeriods(copy.deepcopy(data), periodlength)
    cmap = plt.get_cmap(cmap)
    im = axes.imshow(
        stacked.values.T, interpolation="nearest", vmin=vmin, vmax=vmax, cmap=cmap
    )
    axes.set_aspect("auto")
    axes.set_ylabel("Hour")
    plt.xlabel("Day")

    # fig.subplots_adjust(right=1.1)
    cbar = fig.colorbar(im, ax=axes)
    cbar.set_label("{} [{:~P}]".format(data.name, data.units))

    fig, axes = save_and_show(
        fig, axes, save, show, close, filename, file_format, dpi, axis_off, extent
    )

    return fig, axes


def _plot_poly_collection(
    ax, verts, zs=None, color=None, cmap=None, vmin=None, vmax=None, **kwargs
):
    """
    Args:
        ax:
        verts:
        zs:
        color:
        cmap:
        vmin:
        vmax:
        **kwargs:
    """
    from matplotlib.collections import PolyCollection

    # if None in zs:
    #     zs = None

    # color=None overwrites specified facecolor/edgecolor with default color
    if color is not None:
        kwargs["color"] = color
    import matplotlib as mpl

    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    poly = PolyCollection(verts, **kwargs)
    if zs is not None:
        poly.set_array(np.asarray(zs))
        poly.set_cmap(cmap)
        poly.set_clim(vmin, vmax)

    ax.add_collection3d(poly, zs=zs, zdir="y")
    # ax.autoscale_view()
    return poly


def _plot_surface(ax, x, y, z, cmap=None, **kwargs):
    """
    Args:
        ax:
        x:
        y:
        z:
        cmap:
        **kwargs:
    """
    if cmap is None:
        cmap = cm.gist_earth

    ls = LightSource(270, 45)
    # To use a custom hillshading mode, override the built-in shading and pass
    # in the rgb colors of the shaded surface calculated from "shade".
    rgb = ls.shade(z, cmap=cm.get_cmap(cmap), vert_exag=0.1, blend_mode="soft")
    surf = ax.plot_surface(
        x,
        y,
        z,
        rstride=1,
        cstride=1,
        facecolors=rgb,
        linewidth=0,
        antialiased=False,
        shade=False,
        **kwargs
    )
    return surf


def _plot_contour(ax, x, y, z, cmap=None, **kwargs):
    """
    Args:
        ax:
        x:
        y:
        z:
        cmap:
        **kwargs:
    """
    if cmap is None:
        cmap = cm.gist_earth
    surf = ax.contour3D(x, y, z, 150, cmap=cmap, **kwargs)
    return surf


def _polygon_under_graph(xlist, ylist):
    """Construct the vertex list which defines the polygon filling the space
    under the (xlist, ylist) line graph. Assumes the xs are in ascending order.

    Args:
        xlist:
        ylist:
    """
    return [(xlist[0], 0.0), *zip(xlist, ylist), (xlist[-1], 0.0)]


EnergySeries.plot3d.__doc__ = plot_energyseries.__doc__
EnergySeries.plot2d.__doc__ = plot_energyseries_map.__doc__



################################################################################
# Module: archetypal.template
# Description:
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import collections
import logging as lg
from enum import IntEnum
from functools import reduce

import tabulate
from archetypal import log, IDF, calc_simple_glazing, timeit
from archetypal.template import MaterialLayer, UmiSchedule, UniqueName
from archetypal.template.gas_material import GasMaterial
from archetypal.template.glazing_material import GlazingMaterial
from archetypal.template.umi_base import UmiBase, Unique
from eppy.bunch_subclass import EpBunch


class WindowConstruction(UmiBase, metaclass=Unique):
    """
    $id, AssemblyCarbon, AssemblyCost, AssemblyEnergy, Category, Comments,
    DataSource, DisassemblyCarbon, DisassemblyEnergy, Layers, Name, Type
    """

    def __init__(
        self,
        Category="Double",
        AssemblyCarbon=0,
        AssemblyCost=0,
        AssemblyEnergy=0,
        DisassemblyCarbon=0,
        DisassemblyEnergy=0,
        **kwargs
    ):
        """Initialize a WindowConstruction.

        Args:
            Category (str): "Single", "Double" or "Triple".
            AssemblyCarbon (float): Assembly Embodied Carbon by m2 of
                construction.
            AssemblyCost (float): Assembly cost by m2 of construction.
            AssemblyEnergy (float): Assembly Embodied Energy by m2; of
                construction.
            DisassemblyCarbon (float): Disassembly embodied carbon by m2 of
                construction.
            DisassemblyEnergy (float): Disassembly embodied energy by m2 of
                construction.
            **kwargs: Other keywords passed to the constructor.
        """
        super(WindowConstruction, self).__init__(**kwargs)
        self.Category = Category
        self.DisassemblyEnergy = DisassemblyEnergy
        self.DisassemblyCarbon = DisassemblyCarbon
        self.AssemblyEnergy = AssemblyEnergy
        self.AssemblyCost = AssemblyCost
        self.AssemblyCarbon = AssemblyCarbon
        self.Layers = None

    def __hash__(self):
        return hash((self.__class__.__name__, self.Name, self.DataSource))

    def __eq__(self, other):
        if not isinstance(other, WindowConstruction):
            return False
        else:
            return all(
                [
                    self.Category == other.Category,
                    self.AssemblyCarbon == other.AssemblyCarbon,
                    self.AssemblyCost == other.AssemblyCost,
                    self.AssemblyEnergy == other.AssemblyEnergy,
                    self.DisassemblyCarbon == other.DisassemblyCarbon,
                    self.DisassemblyEnergy == other.DisassemblyEnergy,
                ]
            )

    @classmethod
    def from_json(cls, *args, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        wc = cls(*args, **kwargs)
        layers = kwargs.get("Layers", None)

        # resolve Material objects from ref
        wc.Layers = [
            MaterialLayer(wc.get_ref(layer["Material"]), layer["Thickness"])
            for layer in layers
        ]
        return wc

    @classmethod
    def from_epbunch(cls, Construction, **kwargs):
        """WindowConstruction from idf Construction Name.

        Example:
            >>> import archetypal as ar
            >>> idf = ar.load_idf("myidf")
            >>> construction_name = "Some construction name"
            >>> ar.WindowConstruction.from_epbunch(Name=construction_name,
            >>> idf=idf)

        Args:
            Construction (EpBunch): The Construction epbunch object.
            **kwargs: Other keywords passed to the constructor.
        """
        Name = Construction.Name
        idf = Construction.theidf
        wc = cls(Name=Name, idf=idf, **kwargs)
        wc.Layers = wc.layers()
        catdict = {1: "Single", 2: "Double", 3: "Triple", 4: "Quadruple"}
        wc.Category = catdict[
            len([lyr for lyr in wc.Layers if isinstance(lyr.Material, GlazingMaterial)])
        ]
        return wc

    def to_json(self):
        """Convert class properties to dict"""
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["Layers"] = [layer.to_dict() for layer in self.Layers]
        data_dict["AssemblyCarbon"] = self.AssemblyCarbon
        data_dict["AssemblyCost"] = self.AssemblyCost
        data_dict["AssemblyEnergy"] = self.AssemblyEnergy
        data_dict["DisassemblyCarbon"] = self.DisassemblyCarbon
        data_dict["DisassemblyEnergy"] = self.DisassemblyEnergy
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    def layers(self):
        """Retrieve layers for the WindowConstruction"""
        c = self.idf.getobject("CONSTRUCTION", self.Name)
        layers = []
        for field in c.fieldnames:
            # Loop through the layers from the outside layer towards the
            # indoor layers and get the material they are made of.
            material = c.get_referenced_object(field)
            if material:
                # Create the WindowMaterial:Glazing or the WindowMaterial:Gas
                # and append to the list of layers
                if material.key.upper() == "WindowMaterial:Glazing".upper():
                    material_obj = GlazingMaterial(
                        Conductivity=material.Conductivity,
                        SolarTransmittance=material.Solar_Transmittance_at_Normal_Incidence,
                        SolarReflectanceFront=material.Front_Side_Solar_Reflectance_at_Normal_Incidence,
                        SolarReflectanceBack=material.Back_Side_Solar_Reflectance_at_Normal_Incidence,
                        VisibleTransmittance=material.Visible_Transmittance_at_Normal_Incidence,
                        VisibleReflectanceFront=material.Front_Side_Visible_Reflectance_at_Normal_Incidence,
                        VisibleReflectanceBack=material.Back_Side_Visible_Reflectance_at_Normal_Incidence,
                        IRTransmittance=material.Infrared_Transmittance_at_Normal_Incidence,
                        IREmissivityFront=material.Front_Side_Infrared_Hemispherical_Emissivity,
                        IREmissivityBack=material.Back_Side_Infrared_Hemispherical_Emissivity,
                        DirtFactor=material.Dirt_Correction_Factor_for_Solar_and_Visible_Transmittance,
                        Type="Uncoated",
                        Name=material.Name,
                        Optical=material.Optical_Data_Type,
                        OpticalData=material.Window_Glass_Spectral_Data_Set_Name,
                        idf=self.idf,
                    )

                    material_layer = MaterialLayer(material_obj, material.Thickness)

                elif material.key.upper() == "WindowMaterial:Gas".upper():
                    # Todo: Make gas name generic, like in UmiTemplate Editor
                    material_obj = GasMaterial(
                        Name=material.Gas_Type.upper(), idf=self.idf
                    )
                    material_layer = MaterialLayer(material_obj, material.Thickness)
                elif material.key.upper() == "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM":
                    glass_properties = calc_simple_glazing(
                        material.Solar_Heat_Gain_Coefficient,
                        material.UFactor,
                        material.Visible_Transmittance,
                    )
                    material_obj = GlazingMaterial(
                        **glass_properties, Name=material.Name, idf=self.idf
                    )

                    material_layer = MaterialLayer(
                        material_obj, glass_properties["Thickness"]
                    )
                else:
                    continue

                layers.append(material_layer)
        return layers

    def combine(self, other, weights=None):
        """Append other to self. Return self + other as a new object. For
        now, simply returns self.

        todo:
            - Implement equivalent window layers for constant u-factor.

        """
        return self


class WindowType(IntEnum):
    External = 0
    Internal = 1


class WindowSetting(UmiBase, metaclass=Unique):
    """Window Settings define the various window-related properties of a
    specific :class:`Zone`. Control natural ventilation, shading and airflow
    networks and more using this class. This class serves the same role as the
    ZoneInformation>Windows tab in the UMI TemplateEditor.

    .. image:: ../images/template/zoneinfo-windows.png

    Classmethods:
        The WindowSetting class implements two constructors that are tailored to
        the eppy_ scripting language:

        - :func:`from_construction` and
        - :func:`from_surface`.

    .. _eppy : https://eppy.readthedocs.io/en/latest/
    """

    def __init__(
        self,
        Construction=None,
        OperableArea=0.8,
        AfnWindowAvailability=None,
        AfnDischargeC=0.65,
        AfnTempSetpoint=20,
        IsVirtualPartition=False,
        IsShadingSystemOn=False,
        ShadingSystemAvailabilitySchedule=None,
        ShadingSystemSetpoint=180,
        ShadingSystemTransmittance=0.5,
        ShadingSystemType=0,
        Type=WindowType.External,
        IsZoneMixingOn=False,
        ZoneMixingAvailabilitySchedule=None,
        ZoneMixingDeltaTemperature=2,
        ZoneMixingFlowRate=0.001,
        **kwargs
    ):
        """Initialize a WindowSetting using default values:

        Args:
            Construction (WindowConstruction): The window construction.
            OperableArea (float): The operable window area as a ratio of total
                window area. eg. 0.8 := 80% of the windows area is operable.
            AfnWindowAvailability:
            AfnDischargeC (float): Airflow Network Discharge Coefficient.
                Default = 0.65.
            AfnTempSetpoint (float): Airflow Network Temperature Setpoint.
                Default = 20 degreeC.
            IsVirtualPartition (bool): Virtual Partition.
            IsShadingSystemOn (bool): Shading is used. Default is False.
            ShadingSystemAvailabilitySchedule (UmiSchedule): Shading system
                availability schedule.
            ShadingSystemSetpoint (float): Shading system setpoint in units of
                W/m2. Default = 180 W/m2.
            ShadingSystemTransmittance (float): Shading system transmittance.
                Default = 0.5.
            ShadingSystemType (int): Shading System Type. 0 = ExteriorShade, 1 =
                InteriorShade.
            Type (int):
            IsZoneMixingOn (bool): Zone mixing.
            ZoneMixingAvailabilitySchedule (UmiSchedule): Zone mixing
                availability schedule.
            ZoneMixingDeltaTemperature (float): Zone mixing delta
            ZoneMixingFlowRate (float): Zone mixing flow rate in units of m3/m2.
                Default = 0.001 m3/m2.
            **kwargs: other keywords passed to the constructor.
        """
        super(WindowSetting, self).__init__(**kwargs)

        self.ZoneMixingAvailabilitySchedule = ZoneMixingAvailabilitySchedule
        self.ShadingSystemAvailabilitySchedule = ShadingSystemAvailabilitySchedule
        self.Construction = Construction
        self.AfnWindowAvailability = AfnWindowAvailability
        self.AfnDischargeC = AfnDischargeC
        self.AfnTempSetpoint = AfnTempSetpoint
        self.IsShadingSystemOn = IsShadingSystemOn
        self.IsVirtualPartition = IsVirtualPartition
        self.IsZoneMixingOn = IsZoneMixingOn
        self.OperableArea = OperableArea
        self.ShadingSystemSetpoint = ShadingSystemSetpoint
        self.ShadingSystemTransmittance = ShadingSystemTransmittance
        self.ShadingSystemType = ShadingSystemType
        self.Type = Type  # Todo: Could be deprecated
        self.ZoneMixingDeltaTemperature = ZoneMixingDeltaTemperature
        self.ZoneMixingFlowRate = ZoneMixingFlowRate

    def __add__(self, other):
        return self.combine(other)

    def __repr__(self):
        v_ = [
            (k, v) for k, v in self.__dict__.items() if not isinstance(v, (dict, IDF))
        ]
        header = "{}: <{}>\n".format(self.Name, self.__class__.mro()[0].__name__)
        return header + tabulate.tabulate(v_, tablefmt="plain")

    def __str__(self):
        return repr(self)

    def __hash__(self):
        return hash((self.__class__.__name__, self.Name, self.DataSource))

    def __eq__(self, other):
        if not isinstance(other, WindowSetting):
            return False
        else:
            return all(
                [
                    self.Construction == other.Construction,
                    self.OperableArea == other.OperableArea,
                    self.AfnWindowAvailability == other.AfnWindowAvailability,
                    self.AfnDischargeC == other.AfnDischargeC,
                    self.AfnTempSetpoint == other.AfnTempSetpoint,
                    self.IsVirtualPartition == other.IsVirtualPartition,
                    self.IsShadingSystemOn == other.IsShadingSystemOn,
                    self.ShadingSystemAvailabilitySchedule
                    == other.ShadingSystemAvailabilitySchedule,
                    self.ShadingSystemSetpoint == other.ShadingSystemSetpoint,
                    self.ShadingSystemTransmittance == other.ShadingSystemTransmittance,
                    self.ShadingSystemType == other.ShadingSystemType,
                    self.Type == other.Type,
                    self.IsZoneMixingOn == other.IsZoneMixingOn,
                    self.ZoneMixingAvailabilitySchedule
                    == other.ZoneMixingAvailabilitySchedule,
                    self.ZoneMixingDeltaTemperature == other.ZoneMixingDeltaTemperature,
                    self.ZoneMixingFlowRate == other.ZoneMixingFlowRate,
                ]
            )

    @classmethod
    def generic(cls, idf):
        """Returns a generic window with SHGC=0.704, UFactor=2.703, Tvis=0.786

        Args:
            idf (IDF):
        """
        idf.add_object(
            "WindowMaterial:SimpleGlazingSystem".upper(),
            Name="SimpleWindow:SINGLE PANE HW WINDOW",
            UFactor=2.703,
            Solar_Heat_Gain_Coefficient=0.704,
            Visible_Transmittance=0.786,
            save=False,
        )

        constr = idf.add_object(
            "CONSTRUCTION",
            Name="SINGLE PANE HW WINDOW",
            Outside_Layer="SimpleWindow:SINGLE PANE HW WINDOW",
            save=False,
        )
        return cls.from_construction(Construction=constr)

    @classmethod
    def from_construction(cls, Construction, **kwargs):
        """Make a :class:`WindowSetting` directly from a Construction_ object.

        .. _Construction : https://bigladdersoftware.com/epx/docs/8-9/input
        -output-reference/group-surface-construction-elements.html
        #construction-000

        Examples:
            >>> import archetypal as ar
            >>> # Given an IDF object
            >>> idf = ar.load_idf("idfname")
            >>> construction = idf.getobject('CONSTRUCTION',
            >>>                              'AEDG-SmOffice 1A Window Fixed')
            >>> ar.WindowSetting.from_construction(Name='test_window',
            >>>                    Construction=construction)

        Args:
            Construction (EpBunch): The construction name for this window.
            **kwargs: Other keywords passed to the constructor.

        Returns:
            (windowSetting): The window setting object.
        """
        name = kwargs.pop("Name", Construction.Name + "_Window")
        kwargs["Name"] = name
        w = cls(idf=Construction.theidf, **kwargs)
        w.Construction = WindowConstruction.from_epbunch(Construction)
        w.AfnWindowAvailability = UmiSchedule.constant_schedule(idf=Construction.theidf)
        w.ShadingSystemAvailabilitySchedule = UmiSchedule.constant_schedule(
            idf=Construction.theidf
        )
        w.ZoneMixingAvailabilitySchedule = UmiSchedule.constant_schedule(
            idf=Construction.theidf
        )
        return w

    @classmethod
    def from_surface(cls, surface):
        """Build a WindowSetting object from a FenestrationSurface:Detailed_
        object. This constructor will detect common window constructions and
        shading devices. Supported Shading and Natural Air flow EnergyPlus
        objects are: WindowProperty:ShadingControl_,
        AirflowNetwork:MultiZone:Surface_.

        Important:
            If an EnergyPlus object is not supported, eg.:
            AirflowNetwork:MultiZone:Component:DetailedOpening_, only a warning
            will be issued in the console for the related object instance and
            default values will be automatically used.

        .. _FenestrationSurface:Detailed:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-thermal-zone-description-geometry.html
           #fenestrationsurfacedetailed
        .. _WindowProperty:ShadingControl:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-thermal-zone-description-geometry.html
           #windowpropertyshadingcontrol
        .. _AirflowNetwork:MultiZone:Surface:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-airflow-network.html#airflownetworkmultizonesurface
        .. _AirflowNetwork:MultiZone:Component:DetailedOpening:
           https://bigladdersoftware.com/epx/docs/8-9/input-output-reference
           /group-airflow-network.html
           #airflownetworkmultizonecomponentdetailedopening

        Args:
            surface (EpBunch): The FenestrationSurface:Detailed_ object.

        Returns:
            (WindowSetting): The window setting object.
        """
        if isinstance(surface, EpBunch) and not surface.Surface_Type.upper() == "DOOR":
            construction = surface.get_referenced_object("Construction_Name")
            construction = WindowConstruction.from_epbunch(construction)
            name = surface.Name
            shading_control = surface.get_referenced_object("Shading_Control_Name")
            attr = {}
            if shading_control:
                # a WindowProperty:ShadingControl_ object can be attached to
                # this window
                attr["IsShadingSystemOn"] = True
                if shading_control["Setpoint"] != "":
                    attr["ShadingSystemSetpoint"] = shading_control["Setpoint"]
                shade_mat = shading_control.get_referenced_object(
                    "Shading_Device_Material_Name"
                )
                # get shading transmittance
                if shade_mat:
                    attr["ShadingSystemTransmittance"] = shade_mat[
                        "Visible_Transmittance"
                    ]
                # get shading control schedule
                if shading_control["Shading_Control_Is_Scheduled"].upper() == "YES":
                    name = shading_control["Schedule_Name"]
                    attr["ShadingSystemAvailabilitySchedule"] = UmiSchedule(
                        Name=name, idf=surface.theidf
                    )
                else:
                    # Determine which behavior of control
                    shade_ctrl_type = shading_control["Shading_Control_Type"]
                    if shade_ctrl_type.lower() == "alwaysoff":
                        attr[
                            "ShadingSystemAvailabilitySchedule"
                        ] = UmiSchedule.constant_schedule(
                            idf=surface.theidf, name="AlwaysOff", hourly_value=0
                        )
                    elif shade_ctrl_type.lower() == "alwayson":
                        attr[
                            "ShadingSystemAvailabilitySchedule"
                        ] = UmiSchedule.constant_schedule(idf=surface.theidf)
                    else:
                        log(
                            'Window "{}" uses a  window control type that '
                            'is not supported: "{}". Reverting to '
                            '"AlwaysOn"'.format(name, shade_ctrl_type),
                            lg.WARN,
                        )
                        attr[
                            "ShadingSystemAvailabilitySchedule"
                        ] = UmiSchedule.constant_schedule(idf=surface.theidf)
                # get shading type
                if shading_control["Shading_Type"] != "":
                    mapping = {
                        "InteriorShade": WindowType(1),
                        "ExteriorShade": WindowType(0),
                        "ExteriorScreen": WindowType(0),
                        "InteriorBlind": WindowType(1),
                        "ExteriorBlind": WindowType(0),
                        "BetweenGlassShade": WindowType(0),
                        "BetweenGlassBlind": WindowType(0),
                        "SwitchableGlazing": WindowType(0),
                    }
                    attr["ShadingSystemType"] = mapping[shading_control["Shading_Type"]]
            else:
                # Set default schedules
                attr[
                    "ShadingSystemAvailabilitySchedule"
                ] = UmiSchedule.constant_schedule(idf=surface.theidf)

            # get airflow network
            afn = next(
                iter(
                    surface.getreferingobjs(
                        iddgroups=["Natural Ventilation and Duct Leakage"],
                        fields=["Surface_Name"],
                    )
                ),
                None,
            )
            if afn:
                attr["OperableArea"] = afn.WindowDoor_Opening_Factor_or_Crack_Factor
                leak = afn.get_referenced_object("Leakage_Component_Name")
                name = afn["Venting_Availability_Schedule_Name"]
                if name != "":
                    attr["AfnWindowAvailability"] = UmiSchedule(
                        Name=name, idf=surface.theidf
                    )
                else:
                    attr["AfnWindowAvailability"] = UmiSchedule.constant_schedule(
                        idf=surface.theidf
                    )
                name = afn[
                    "Ventilation_Control_Zone_Temperature_Setpoint_Schedule_Name"
                ]
                if name != "":
                    attr["AfnTempSetpoint"] = UmiSchedule(
                        Name=name, idf=surface.theidf
                    ).mean
                else:
                    pass  # uses default

                if (
                    leak.key.upper()
                    == "AIRFLOWNETWORK:MULTIZONE:SURFACE:EFFECTIVELEAKAGEAREA"
                ):
                    attr["AfnDischargeC"] = leak["Discharge_Coefficient"]
                elif (
                    leak.key.upper()
                    == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:HORIZONTALOPENING"
                ):
                    log(
                        '"{}" is not fully supported. Rerverting to '
                        'defaults for object "{}"'.format(
                            leak.key, cls.mro()[0].__name__
                        ),
                        lg.WARNING,
                    )
                elif leak.key.upper() == "AIRFLOWNETWORK:MULTIZONE:SURFACE:CRACK":
                    log(
                        '"{}" is not fully supported. Rerverting to '
                        'defaults for object "{}"'.format(
                            leak.key, cls.mro()[0].__name__
                        ),
                        lg.WARNING,
                    )
                elif (
                    leak.key.upper()
                    == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:DETAILEDOPENING"
                ):
                    log(
                        '"{}" is not fully supported. Rerverting to '
                        'defaults for object "{}"'.format(
                            leak.key, cls.mro()[0].__name__
                        ),
                        lg.WARNING,
                    )
                elif (
                    leak.key.upper()
                    == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:ZONEEXHAUSTFAN"
                ):
                    log(
                        '"{}" is not fully supported. Rerverting to '
                        'defaults for object "{}"'.format(
                            leak.key, cls.mro()[0].__name__
                        ),
                        lg.WARNING,
                    )
                elif (
                    leak.key.upper()
                    == "AIRFLOWNETWORK:MULTIZONE:COMPONENT:SIMPLEOPENING"
                ):
                    log(
                        '"{}" is not fully supported. Rerverting to '
                        'defaults for object "{}"'.format(
                            leak.key, cls.mro()[0].__name__
                        ),
                        lg.WARNING,
                    )
            else:
                attr["AfnWindowAvailability"] = UmiSchedule.constant_schedule(
                    idf=surface.theidf
                )
            # Zone Mixing
            attr["ZoneMixingAvailabilitySchedule"] = UmiSchedule.constant_schedule(
                idf=surface.theidf
            )
            w = cls(
                Name=name,
                Construction=construction,
                idf=surface.theidf,
                Category=surface.theidf.building_name(use_idfname=True),
                **attr
            )
            return w

    @classmethod
    @timeit
    def from_zone(cls, zone):
        """Iterate over the zone subsurfaces and create a window object. If more
        than one window is created, use reduce to combine them together.

        Args:
            zone (Zone): The Zone object from which the WindowSetting is
                created.

        Returns:
            WindowSetting: The WindowSetting object for this zone.
        """
        window_sets = []

        for surf in zone._zonesurfaces:
            # skip internalmass objects since they don't have windows.
            if surf.key.lower() != "internalmass":
                for subsurf in surf.subsurfaces:
                    # For each subsurface, create a WindowSetting object
                    # using the `from_surface` constructor.
                    if subsurf.Surface_Type.lower() == "window":
                        window_sets.append(cls.from_surface(subsurf))

        if window_sets:
            # if one or more window has been created, reduce. Using reduce on
            # a len==1 list, will simply return the object.
            from operator import add

            return reduce(add, window_sets)
        else:
            # no window found, probably a core zone, return None.
            return None

    def combine(self, other, weights=None):
        """Append other to self. Return self + other as a new object.

        Args:
            other (WindowSetting): The other OpaqueMaterial object
            weights (list-like, optional): A list-like object of len 2. If None,
                equal weights are used.

        Returns:
            WindowSetting: A new combined object made of self + other.
        """
        if self is None:
            return other
        if other is None:
            return self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        if not weights:
            log(
                'using 1 as weighting factor in "{}" '
                "combine.".format(self.__class__.__name__)
            )
            weights = [1.0, 1.0]
        meta = self._get_predecessors_meta(other)
        new_attr = dict(
            Construction=self.Construction.combine(other.Construction, weights),
            AfnDischargeC=self._float_mean(other, "AfnDischargeC", weights),
            AfnTempSetpoint=self._float_mean(other, "AfnTempSetpoint", weights),
            AfnWindowAvailability=self.AfnWindowAvailability.combine(
                other.AfnWindowAvailability, weights),
            IsShadingSystemOn=any([self.IsShadingSystemOn, other.IsShadingSystemOn]),
            IsVirtualPartition=any([self.IsVirtualPartition, other.IsVirtualPartition]),
            IsZoneMixingOn=any([self.IsZoneMixingOn, other.IsZoneMixingOn]),
            OperableArea=self._float_mean(other, "OperableArea", weights),
            ShadingSystemSetpoint=self._float_mean(
                other, "ShadingSystemSetpoint", weights
            ),
            ShadingSystemTransmittance=self._float_mean(
                other, "ShadingSystemTransmittance", weights
            ),
            ShadingSystemType=self.ShadingSystemType
            if self.IsShadingSystemOn
            else other.ShadingSystemType,
            ZoneMixingDeltaTemperature=self._float_mean(
                other, "ZoneMixingDeltaTemperature", weights
            ),
            ZoneMixingFlowRate=self._float_mean(other, "ZoneMixingFlowRate", weights),
            ZoneMixingAvailabilitySchedule=self.ZoneMixingAvailabilitySchedule.combine(
                other.ZoneMixingAvailabilitySchedule, weights),
            ShadingSystemAvailabilitySchedule=self.ShadingSystemAvailabilitySchedule
                .combine(
                other.ShadingSystemAvailabilitySchedule, weights),
        )
        new_obj = self.__class__(**meta, **new_attr)
        new_obj._predecessors.extend(self._predecessors + other._predecessors)
        return new_obj

    def to_json(self):
        """Convert class properties to dict"""
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["AfnDischargeC"] = self.AfnDischargeC
        data_dict["AfnTempSetpoint"] = self.AfnTempSetpoint
        data_dict["AfnWindowAvailability"] = self.AfnWindowAvailability.to_dict()
        data_dict["Construction"] = {"$ref": str(self.Construction.id)}
        data_dict["IsShadingSystemOn"] = self.IsShadingSystemOn
        data_dict["IsVirtualPartition"] = self.IsVirtualPartition
        data_dict["IsZoneMixingOn"] = self.IsZoneMixingOn
        data_dict["OperableArea"] = self.OperableArea
        data_dict[
            "ShadingSystemAvailabilitySchedule"
        ] = self.ShadingSystemAvailabilitySchedule.to_dict()
        data_dict["ShadingSystemSetpoint"] = self.ShadingSystemSetpoint
        data_dict["ShadingSystemTransmittance"] = self.ShadingSystemTransmittance
        data_dict["ShadingSystemType"] = self.ShadingSystemType
        data_dict["Type"] = self.Type
        data_dict[
            "ZoneMixingAvailabilitySchedule"
        ] = self.ZoneMixingAvailabilitySchedule.to_dict()
        data_dict["ZoneMixingDeltaTemperature"] = self.ZoneMixingDeltaTemperature
        data_dict["ZoneMixingFlowRate"] = self.ZoneMixingFlowRate
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    @classmethod
    def from_json(cls, *args, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        w = cls(*args, **kwargs)

        ref = kwargs.get("AfnWindowAvailability", None)
        w.AfnWindowAvailability = w.get_ref(ref)
        ref = kwargs.get("Construction", None)
        w.Construction = w.get_ref(ref)
        ref = kwargs.get("ShadingSystemAvailabilitySchedule", None)
        w.ShadingSystemAvailabilitySchedule = w.get_ref(ref)
        ref = kwargs.get("ZoneMixingAvailabilitySchedule", None)
        w.ZoneMixingAvailabilitySchedule = w.get_ref(ref)
        return w



################################################################################
# Module: archetypal.template
# Description:
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import collections
import math

import numpy as np
from archetypal import float_round, ReportData, log, timeit, settings
from archetypal.template import UmiBase, Unique, UmiSchedule, UniqueName


class ZoneConditioning(UmiBase, metaclass=Unique):
    """HVAC settings for the zone

    .. image:: ../images/template/zoninfo-conditioning.png
    """

    def __init__(
        self,
        CoolingCoeffOfPerf=1,
        CoolingLimitType="NoLimit",
        CoolingSetpoint=26,
        CoolingSchedule=None,
        EconomizerType="NoEconomizer",
        HeatRecoveryEfficiencyLatent=0.65,
        HeatRecoveryEfficiencySensible=0.7,
        HeatRecoveryType="None",
        HeatingCoeffOfPerf=1,
        HeatingLimitType="NoLimit",
        HeatingSetpoint=20,
        HeatingSchedule=None,
        IsCoolingOn=True,
        IsHeatingOn=True,
        IsMechVentOn=True,
        MaxCoolFlow=100,
        MaxCoolingCapacity=100,
        MaxHeatFlow=100,
        MaxHeatingCapacity=100,
        MinFreshAirPerArea=0,
        MinFreshAirPerPerson=0.00944,
        MechVentSchedule=None,
        **kwargs,
    ):
        """Initialize a new :class:`ZoneConditioning` object.

        Args:
            CoolingCoeffOfPerf (float): Performance factor of the cooling system.
                This value is used to calculate the total cooling energy use by
                dividing the cooling load by the COP. The COP of the zone shared with all zones
                and refers to the COP of the entire building.
            CoolingLimitType (str): The input must be either LimitFlowRate,
                LimitCapacity, LimitFlowRateAndCapacity or NoLimit.
            CoolingSetpoint (float): The temperature above which the zone heating is
                turned on. Here, we take the mean value over the year.
            CoolingSchedule (UmiSchedule): The availability schedule for space
                cooling in this zone. If the value is 0, cooling is not available,
                and cooling is not supplied to the zone.
            EconomizerType (str): Specifies if there is an outdoor air
                economizer. The choices are: NoEconomizer, DifferentialDryBulb,
                or DifferentialEnthalpy. For the moment, the EconomizerType is
                applied for the entire building (every zone with the same
                EconomizerType). Moreover, since UMI does not support all Economizer
                Types, some assumptions are made:

                - If 'NoEconomizer' in EnergyPlus, EconomizerType='NoEconomizer'
                - If 'DifferentialEnthalpy' in EnergyPlus,EconomizerType =
                  'DifferentialEnthalpy'
                - If 'DifferentialDryBulb' in EnergyPlus, EconomizerType =
                  'DifferentialDryBulb'
                - If 'FixedDryBulb' in EnergyPlus, EconomizerType =
                  'DifferentialDryBulb'
                - If 'FixedEnthalpy' in EnergyPlus, EconomizerType =
                  'DifferentialEnthalpy'
                - If 'ElectronicEnthalpy' in EnergyPlus, EconomizerType =
                  'DifferentialEnthalpy'
                - If 'FixedDewPointAndDryBulb' in EnergyPlus, EconomizerType =
                  'DifferentialDryBulb'
                - If 'DifferentialDryBulbAndEnthalpy' in EnergyPlus,
                  EconomizerType = 'DifferentialEnthalpy'
            HeatRecoveryEfficiencyLatent (float): The latent heat recovery
                effectiveness, where effectiveness is defined as the change in
                supply humidity ratio divided by the difference in entering
                supply and relief air humidity ratios. The default is 0.65.

                - If the HeatExchanger is an AirToAir FlatPlate,
                  HeatRecoveryEfficiencyLatent = HeatRecoveryEfficiencySensible
                  - 0.05
                - If the HeatExchanger is an AirToAir SensibleAndLatent, we
                  suppose that HeatRecoveryEfficiencyLatent = Latent
                  Effectiveness at 100% Heating Air Flow
                - If the HeatExchanger is a Desiccant BalancedFlow, we use the
                  default value for the efficiency (=0.65).
            HeatRecoveryEfficiencySensible (float): The sensible heat recovery
                effectiveness, where effectiveness is defined as the change in
                supply temperature divided by the difference in entering supply
                and relief air temperatures. The default is 0.70.

                - If the HeatExchanger is an AirToAir FlatPlate,
                  HeatRecoveryEfficiencySensible = (Supply Air Outlet T°C -
                  Supply Air Inlet T°C)/(Secondary Air Inlet T°C - Supply Air
                  Inlet T°C)
                - If the HeatExchanger is an AirToAir SensibleAndLatent, we
                  suppose that HeatRecoveryEfficiencySensible = Sensible
                  Effectiveness at 100% Heating Air Flow
                - If the HeatExchanger is a Desiccant BalancedFlow, we use the
                  default value for the efficiency (=0.70)
            HeatRecoveryType (str): Select from *None*, *Sensible*, or
                *Enthalpy*. None means that there is no heat recovery. Sensible
                means that there is sensible heat recovery whenever the zone
                exhaust air temperature is more favorable than the outdoor air
                temperature. Enthalpy means that there is latent and sensible
                heat recovery whenever the zone exhaust air enthalpy is more
                favorable than the outdoor air enthalpy. The default is None
            HeatingCoeffOfPerf (float): Efficiency of heating system. The COP is
                of each zone is equal, and refer to the COP of the entire
                building.
            HeatingLimitType (str): The input must be either LimitFlowRate,
                LimitCapacity, LimitFlowRateAndCapacity or NoLimit.
            HeatingSetpoint (float): The temperature below which zone heating is
                turned on. Here, we take the mean value over the year.
            HeatingSchedule (UmiSchedule): The availability schedule for space
                heating in this zone. If the value is 0, heating is not available,
                and heating is not supplied to the zone.
            IsCoolingOn (bool): Whether or not cooling is available.
            IsHeatingOn (bool): Whether or not heating is available.
            IsMechVentOn (bool): If True, an outdoor air quantity for use by the
                model is calculated.
            MaxCoolFlow (float): The maximum cooling supply air flow rate in
                cubic meters per second if Cooling Limit is set to LimitFlowRate
                or LimitFlowRateAndCapacity
            MaxCoolingCapacity (float): The maximum allowed total (sensible plus
                latent) cooling capacity in Watts per square meter.
            MaxHeatFlow (float): The maximum heating supply air flow rate in
                cubic meters per second if heating limit is set to LimitFlowRate
                or LimitFlowRateAndCapacity
            MaxHeatingCapacity (float): The maximum allowed sensible heating
                capacity in Watts if Heating Limit is set to LimitCapacity or
                LimitFlowRateAndCapacity
            MinFreshAirPerArea (flaot): The design outdoor air volume flow rate
                per square meter of floor area (units are m3/s-m2). This input
                is used if Outdoor Air Method is Flow/Area, Sum or Maximum
            MinFreshAirPerPerson (float): The design outdoor air volume flow
                rate per person for this zone in cubic meters per second per
                person. The default is 0.00944 (20 cfm per person).
            MechVentSchedule (UmiSchedule): The availability schedule of the
                mechanical ventilation. If the value is 0, the mechanical ventilation is
                not available and air flow is not requested.
            **kwargs: Other arguments passed to the base class
                :class:`archetypal.template.UmiBase`
        """
        super(ZoneConditioning, self).__init__(**kwargs)
        self.MechVentSchedule = MechVentSchedule
        self.HeatingSchedule = HeatingSchedule
        self.CoolingSchedule = CoolingSchedule
        self.CoolingCoeffOfPerf = CoolingCoeffOfPerf
        self.CoolingLimitType = CoolingLimitType
        self.CoolingSetpoint = CoolingSetpoint
        self.EconomizerType = EconomizerType
        self.HeatRecoveryEfficiencyLatent = HeatRecoveryEfficiencyLatent
        self.HeatRecoveryEfficiencySensible = HeatRecoveryEfficiencySensible
        self.HeatRecoveryType = HeatRecoveryType
        self.HeatingCoeffOfPerf = HeatingCoeffOfPerf
        self.HeatingLimitType = HeatingLimitType
        self.HeatingSetpoint = HeatingSetpoint
        self.IsCoolingOn = IsCoolingOn
        self.IsHeatingOn = IsHeatingOn
        self.IsMechVentOn = IsMechVentOn
        self.MaxCoolFlow = MaxCoolFlow
        self.MaxCoolingCapacity = MaxCoolingCapacity
        self.MaxHeatFlow = MaxHeatFlow
        self.MaxHeatingCapacity = MaxHeatingCapacity
        self.MinFreshAirPerArea = MinFreshAirPerArea
        self.MinFreshAirPerPerson = MinFreshAirPerPerson

        self._belongs_to_zone = kwargs.get("zone", None)

    def __add__(self, other):
        return self.combine(other)

    def __hash__(self):
        return hash((self.__class__.__name__, self.Name, self.DataSource))

    def __eq__(self, other):
        if not isinstance(other, ZoneConditioning):
            return False
        else:
            return all(
                [
                    self.CoolingCoeffOfPerf == other.CoolingCoeffOfPerf,
                    self.CoolingLimitType == other.CoolingLimitType,
                    self.CoolingSetpoint == other.CoolingSetpoint,
                    self.CoolingSchedule == other.CoolingSchedule,
                    self.EconomizerType == other.EconomizerType,
                    self.HeatRecoveryEfficiencyLatent
                    == other.HeatRecoveryEfficiencyLatent,
                    self.HeatRecoveryEfficiencySensible
                    == other.HeatRecoveryEfficiencySensible,
                    self.HeatRecoveryType == other.HeatRecoveryType,
                    self.HeatingCoeffOfPerf == other.HeatingCoeffOfPerf,
                    self.HeatingLimitType == other.HeatingLimitType,
                    self.HeatingSetpoint == other.HeatingSetpoint,
                    self.HeatingSchedule == other.HeatingSchedule,
                    self.IsCoolingOn == other.IsCoolingOn,
                    self.IsHeatingOn == other.IsHeatingOn,
                    self.IsMechVentOn == other.IsMechVentOn,
                    self.MaxCoolFlow == other.MaxCoolFlow,
                    self.MaxCoolingCapacity == other.MaxCoolingCapacity,
                    self.MaxHeatFlow == other.MaxHeatFlow,
                    self.MaxHeatingCapacity == other.MaxHeatingCapacity,
                    self.MinFreshAirPerArea == other.MinFreshAirPerArea,
                    self.MinFreshAirPerPerson == other.MinFreshAirPerPerson,
                    self.MechVentSchedule == other.MechVentSchedule,
                ]
            )

    @classmethod
    def from_json(cls, *args, **kwargs):
        """
        Args:
            *args:
            **kwargs:
        """
        zc = cls(*args, **kwargs)

        cool_schd = kwargs.get("CoolingSchedule", None)
        zc.CoolingSchedule = zc.get_ref(cool_schd)
        heat_schd = kwargs.get("HeatingSchedule", None)
        zc.HeatingSchedule = zc.get_ref(heat_schd)
        mech_schd = kwargs.get("MechVentSchedule", None)
        zc.MechVentSchedule = zc.get_ref(mech_schd)
        return zc

    def to_json(self):
        """Convert class properties to dict"""
        data_dict = collections.OrderedDict()

        data_dict["$id"] = str(self.id)
        data_dict["CoolingSchedule"] = self.CoolingSchedule.to_dict()
        data_dict["CoolingCoeffOfPerf"] = self.CoolingCoeffOfPerf
        data_dict["CoolingSetpoint"] = (
            self.CoolingSetpoint if not math.isnan(self.CoolingSetpoint) else 26
        )
        data_dict["CoolingLimitType"] = self.CoolingLimitType
        data_dict["EconomizerType"] = self.EconomizerType
        data_dict["HeatingCoeffOfPerf"] = self.HeatingCoeffOfPerf
        data_dict["HeatingLimitType"] = self.HeatingLimitType
        data_dict["HeatingSchedule"] = self.HeatingSchedule.to_dict()
        data_dict["HeatingSetpoint"] = (
            self.HeatingSetpoint if not math.isnan(self.HeatingSetpoint) else 20
        )
        data_dict["HeatRecoveryEfficiencyLatent"] = self.HeatRecoveryEfficiencyLatent
        data_dict[
            "HeatRecoveryEfficiencySensible"
        ] = self.HeatRecoveryEfficiencySensible
        data_dict["HeatRecoveryType"] = self.HeatRecoveryType
        data_dict["IsCoolingOn"] = self.IsCoolingOn
        data_dict["IsHeatingOn"] = self.IsHeatingOn
        data_dict["IsMechVentOn"] = self.IsMechVentOn
        data_dict["MaxCoolFlow"] = self.MaxCoolFlow
        data_dict["MaxCoolingCapacity"] = self.MaxCoolingCapacity
        data_dict["MaxHeatFlow"] = self.MaxHeatFlow
        data_dict["MaxHeatingCapacity"] = self.MaxHeatingCapacity
        data_dict["MechVentSchedule"] = self.MechVentSchedule.to_dict()
        data_dict["MinFreshAirPerArea"] = self.MinFreshAirPerArea
        data_dict["MinFreshAirPerPerson"] = self.MinFreshAirPerPerson
        data_dict["Category"] = self.Category
        data_dict["Comments"] = self.Comments
        data_dict["DataSource"] = self.DataSource
        data_dict["Name"] = UniqueName(self.Name)

        return data_dict

    @classmethod
    @timeit
    def from_zone(cls, zone):
        """
        Args:
            zone (archetypal.template.zone.Zone): zone to gets information from
        """
        # First create placeholder object.
        name = zone.Name + "_ZoneConditioning"
        z_cond = cls(
            Name=name,
            zone=zone,
            idf=zone.idf,
            Category=zone.idf.building_name(use_idfname=True),
        )

        z_cond._set_thermostat_setpoints(zone)

        z_cond._set_zone_cops(zone)

        z_cond._set_heat_recovery(zone)

        z_cond._set_mechanical_ventilation(zone)

        z_cond._set_economizer(zone)

        return z_cond

    def _set_economizer(self, zone):
        """Set economizer parameters

        Todo:
            - Here EconomizerType is for the entire building, try to do it for
              each zone.
            - Fix typo in DifferentialEnthalpy (extra h) when issue is resolved
              at Basilisk project:
              https://github.com/MITSustainableDesignLab/basilisk/issues/32

        Args:
            zone (Zone): The zone object.
        """
        # Economizer
        controllers_in_idf = zone.idf.idfobjects["Controller:OutdoorAir".upper()]
        self.EconomizerType = "NoEconomizer"  # default value

        for object in controllers_in_idf:
            if object.Economizer_Control_Type == "NoEconomizer":
                self.EconomizerType = "NoEconomizer"
            elif object.Economizer_Control_Type == "DifferentialEnthalphy":
                self.EconomizerType = "DifferentialEnthalphy"
            elif object.Economizer_Control_Type == "DifferentialDryBulb":
                self.EconomizerType = "DifferentialDryBulb"
            elif object.Economizer_Control_Type == "FixedDryBulb":
                self.EconomizerType = "DifferentialDryBulb"
            elif object.Economizer_Control_Type == "FixedEnthalpy":
                self.EconomizerType = "DifferentialEnthalphy"
            elif object.Economizer_Control_Type == "ElectronicEnthalpy":
                self.EconomizerType = "DifferentialEnthalphy"
            elif object.Economizer_Control_Type == "FixedDewPointAndDryBulb":
                self.EconomizerType = "DifferentialDryBulb"
            elif object.Economizer_Control_Type == "DifferentialDryBulbAndEnthalpy":
                self.EconomizerType = "DifferentialEnthalphy"

    def _set_mechanical_ventilation(self, zone):
        """Iterate on 'Controller:MechanicalVentilation' objects to find the
        'DesignSpecifactionOutdoorAirName' for the zone.

        Todo:
            - Get mechanical sizing by zone.

        Args:
            zone (Zone): The zone object.
        """
        if not zone.idf.idfobjects["Controller:MechanicalVentilation".upper()]:
            IsMechVentOn = False
            MinFreshAirPerArea = 0  # use defaults
            MinFreshAirPerPerson = 0
            MechVentSchedule = UmiSchedule.constant_schedule(idf=zone.idf)
        else:
            design_spe_outdoor_air_name = ""
            for object in zone.idf.idfobjects[
                "Controller:MechanicalVentilation".upper()
            ]:
                if zone.Name in object.fieldvalues:
                    indice_zone = [
                        k for k, s in enumerate(object.fieldvalues) if s == zone.Name
                    ][0]
                    design_spe_outdoor_air_name = object.fieldvalues[indice_zone + 1]

                    if object.Availability_Schedule_Name != "":
                        MechVentSchedule = UmiSchedule(
                            Name=object.Availability_Schedule_Name, idf=zone.idf
                        )
                    else:
                        MechVentSchedule = UmiSchedule.constant_schedule(idf=zone.idf)
                    break
            # If 'DesignSpecifactionOutdoorAirName', MechVent is ON, and gets
            # the minimum fresh air (per person and area)
            if design_spe_outdoor_air_name != "":
                IsMechVentOn = True
                design_spe_outdoor_air = zone.idf.getobject(
                    "DesignSpecification:OutdoorAir".upper(),
                    design_spe_outdoor_air_name,
                )
                MinFreshAirPerPerson = (
                    design_spe_outdoor_air.Outdoor_Air_Flow_per_Person
                )
                # If MinFreshAirPerPerson NOT a number, MinFreshAirPerPerson=0
                try:
                    MinFreshAirPerPerson = float(MinFreshAirPerPerson)
                except:
                    MinFreshAirPerPerson = 0
                MinFreshAirPerArea = (
                    design_spe_outdoor_air.Outdoor_Air_Flow_per_Zone_Floor_Area
                )
                # If MinFreshAirPerArea NOT a number, MinFreshAirPerArea=0
                try:
                    MinFreshAirPerArea = float(MinFreshAirPerArea)
                except:
                    MinFreshAirPerArea = 0
            else:
                IsMechVentOn = False
                MinFreshAirPerPerson = 0
                MinFreshAirPerArea = 0
                MechVentSchedule = UmiSchedule.constant_schedule(idf=zone.idf)

        self.IsMechVentOn = IsMechVentOn
        self.MinFreshAirPerArea = MinFreshAirPerArea
        self.MinFreshAirPerPerson = MinFreshAirPerPerson
        self.MechVentSchedule = MechVentSchedule

    def _set_zone_cops(self, zone):
        """
        Todo:
            - Make this method zone-independent.

        Args:
            zone (Zone):
        """
        # COPs (heating and cooling)

        # Heating
        heating_meters = (
            "Heating:Electricity",
            "Heating:Gas",
            "Heating:DistrictHeating",
        )
        heating_cop = self._get_cop(
            zone,
            energy_in_list=heating_meters,
            energy_out_variable_name=(
                "Air System Total Heating Energy",
                "Zone Ideal Loads Zone Total Heating Energy",
            ),
        )
        # Cooling
        cooling_meters = (
            "Cooling:Electricity",
            "Cooling:Gas",
            "Cooling:DistrictCooling",
        )
        cooling_cop = self._get_cop(
            zone,
            energy_in_list=cooling_meters,
            energy_out_variable_name=(
                "Air System Total Cooling Energy",
                "Zone Ideal Loads Zone Total Cooling Energy",
            ),
        )

        # Capacity limits (heating and cooling)
        zone_size = zone.sql["ZoneSizes"][
            zone.sql["ZoneSizes"]["ZoneName"] == zone.Name.upper()
        ]
        # Heating
        HeatingLimitType, heating_cap, heating_flow = self._get_design_limits(
            zone, zone_size, load_name="Heating"
        )
        # Cooling
        CoolingLimitType, cooling_cap, cooling_flow = self._get_design_limits(
            zone, zone_size, load_name="Cooling"
        )

        self.HeatingLimitType = HeatingLimitType
        self.MaxHeatingCapacity = heating_cap
        self.MaxHeatFlow = heating_flow
        self.CoolingLimitType = CoolingLimitType
        self.MaxCoolingCapacity = cooling_cap
        self.MaxCoolFlow = cooling_flow

        self.CoolingCoeffOfPerf = cooling_cop
        self.HeatingCoeffOfPerf = heating_cop

        # If cop calc == infinity, COP = 1 because we need a value in json file.
        if heating_cop == float("infinity"):
            self.HeatingCoeffOfPerf = 1
        if cooling_cop == float("infinity"):
            self.CoolingCoeffOfPerf = 1

    def _set_thermostat_setpoints(self, zone):
        """Sets the thermostat settings and schedules for this zone.

        Thermostat Setpoints:
            - CoolingSetpoint (float):
            - HeatingSetpoint (float):
            - HeatingSchedule (UmiSchedule):
            - CoolingSchedule (UmiSchedule):

        Args:
            zone (Zone): The zone object.
        """
        # Set Thermostat set points
        # Heating and Cooling set points and schedules
        self.HeatingSetpoint, self.HeatingSchedule, self.CoolingSetpoint, self.CoolingSchedule = self._get_setpoint_and_scheds(
            zone
        )

        # If HeatingSetpoint == nan, means there is no heat or cold input,
        # therefore system is off.
        if math.isnan(self.HeatingSetpoint):
            self.IsHeatingOn = False
        else:
            self.IsHeatingOn = True
        if math.isnan(self.CoolingSetpoint):
            self.IsCoolingOn = False
        else:
            self.IsCoolingOn = True

    def _set_heat_recovery(self, zone):
        """Sets the heat recovery parameters for this zone.

        Heat Recovery Parameters:
            - HeatRecoveryEfficiencyLatent (float): The latent heat recovery
              effectiveness.
            - HeatRecoveryEfficiencySensible (float): The sensible heat recovery
              effectiveness.
            - HeatRecoveryType (str): 'None', 'Sensible' or 'Enthalpy'.
            - comment (str): A comment to append to the class comment attribute.

        Args:
            zone (Zone): The Zone object.
        """
        from itertools import chain

        # Todo: Implement loop that detects HVAC linked to Zone; than parse heat
        #  recovery. Needs to happen when a zone has a ZoneHVAC:IdealLoadsAirSystem
        # connections = zone._epbunch.getreferingobjs(
        #     iddgroups=["Zone HVAC Equipment Connections"], fields=["Zone_Name"]
        # )
        # nodes = [
        #     con.get_referenced_object("Zone_Air_Inlet_Node_or_NodeList_Name")
        #     for con in connections
        # ]

        # get possible heat recovery objects from idd
        heat_recovery_objects = zone.idf.getiddgroupdict()["Heat Recovery"]

        # get possible heat recovery objects from this idf
        heat_recovery_in_idf = list(
            chain.from_iterable(
                zone.idf.idfobjects[key.upper()] for key in heat_recovery_objects
            )
        )

        # Set defaults
        HeatRecoveryEfficiencyLatent = 0.7
        HeatRecoveryEfficiencySensible = 0.65
        HeatRecoveryType = "None"
        comment = ""

        # iterate over those objects. If the list is empty, it will simply pass.
        for object in heat_recovery_in_idf:

            if object.key.upper() == "HeatExchanger:AirToAir:FlatPlate".upper():
                # Do HeatExchanger:AirToAir:FlatPlate

                nsaot = object.Nominal_Supply_Air_Outlet_Temperature
                nsait = object.Nominal_Supply_Air_Inlet_Temperature
                n2ait = object.Nominal_Secondary_Air_Inlet_Temperature
                HeatRecoveryEfficiencySensible = (nsaot - nsait) / (n2ait - nsait)
                # Hypotheses: HeatRecoveryEfficiencySensible - 0.05
                HeatRecoveryEfficiencyLatent = HeatRecoveryEfficiencySensible - 0.05
                HeatRecoveryType = "Enthalpy"
                comment = (
                    "HeatRecoveryEfficiencySensible was calculated "
                    "using this formula: (Supply Air Outlet T°C -; "
                    "Supply Air Inlet T°C)/(Secondary Air Inlet T°C - "
                    "Supply Air Inlet T°C)"
                )

            elif (
                object.key.upper() == "HeatExchanger:AirToAir:SensibleAndLatent".upper()
            ):
                # Do HeatExchanger:AirToAir:SensibleAndLatent

                HeatRecoveryEfficiencyLatent, HeatRecoveryEfficiencySensible = self._get_recoverty_effectiveness(
                    object, zone
                )
                HeatRecoveryType = "Enthalpy"

                comment = (
                    "HeatRecoveryEfficiencies were calculated using "
                    "simulation hourly values and averaged. Only values"
                    " > 0 were used in the average calculation."
                )

            elif object.key.upper() == "HeatExchanger:Desiccant:BalancedFlow".upper():
                # Do HeatExchanger:Dessicant:BalancedFlow
                # Use default values
                HeatRecoveryEfficiencyLatent = 0.7
                HeatRecoveryEfficiencySensible = 0.65
                HeatRecoveryType = "Enthalpy"

            elif (
                object.key.upper() == "HeatExchanger:Desiccant:BalancedFlow"
                ":PerformanceDataType1".upper()
            ):
                # This is not an actual HeatExchanger, pass
                pass
            else:
                msg = 'Heat exchanger object "{}" is not ' "implemented".format(object)
                raise NotImplementedError(msg)

        self.HeatRecoveryEfficiencyLatent = HeatRecoveryEfficiencyLatent
        self.HeatRecoveryEfficiencySensible = HeatRecoveryEfficiencySensible
        self.HeatRecoveryType = HeatRecoveryType
        self.Comments += comment

    @staticmethod
    def _get_recoverty_effectiveness(object, zone):
        """
        Args:
            object:
            zone:
        """
        rd = ReportData.from_sql_dict(zone.sql)
        effectiveness = (
            rd.filter_report_data(
                name=(
                    "Heat Exchanger Sensible Effectiveness",
                    "Heat Exchanger Latent Effectiveness",
                )
            )
            .loc[lambda x: x.Value > 0]
            .groupby(["KeyValue", "Name"])
            .Value.mean()
            .unstack(level=-1)
        )
        HeatRecoveryEfficiencySensible = effectiveness.loc[
            object.Name.upper(), "Heat Exchanger Sensible " "Effectiveness"
        ]
        HeatRecoveryEfficiencyLatent = effectiveness.loc[
            object.Name.upper(), "Heat Exchanger Latent Effectiveness"
        ]
        return HeatRecoveryEfficiencyLatent, HeatRecoveryEfficiencySensible

    @classmethod
    def _get_design_limits(cls, zone, zone_size, load_name):
        """Gets design limits for heating and cooling systems

        Args:
            zone (archetypal.template.zone.Zone): zone to gets information from
            zone_size (df): Dataframe from the sql EnergyPlus outpout, with the
                sizing of the heating and cooling systems
            load_name (str): 'Heating' or 'Cooling' depending on what system we
                want to characterize
        """
        try:
            cap = round(
                zone_size[zone_size["LoadType"] == load_name]["UserDesLoad"].values[0]
                / zone.area,
                3,
            )
            flow = (
                zone_size[zone_size["LoadType"] == load_name]["UserDesFlow"].values[0]
                / zone.area
            )
            LimitType = "LimitFlowRateAndCapacity"
        except:
            cap = 100
            flow = 100
            LimitType = "NoLimit"
        return LimitType, cap, flow

    @classmethod
    def _get_cop(cls, zone, energy_in_list, energy_out_variable_name):
        """Calculates COP for heating or cooling systems

        Args:
            zone (archetypal.template.zone.Zone): zone to gets information from
            energy_in_list (str or tuple): list of the energy sources for a
                system (e.g. [Heating:Electricity, Heating:Gas] for heating
                system)
            energy_out_variable_name (str or tuple): Name of the output in the
                sql for the energy given to the zone from the system (e.g. 'Air
                System Total Heating Energy')
        """
        from archetypal import ReportData

        rd = ReportData.from_sql_dict(zone.sql)
        energy_out = rd.filter_report_data(name=tuple(energy_out_variable_name))
        energy_in = rd.filter_report_data(name=tuple(energy_in_list))

        # zone_to_hvac = {zone.Zone_Name: [
        #     zone.get_referenced_object(
        #         'Zone_Conditioning_Equipment_List_Name
        #         ').get_referenced_object(
        #         fieldname) for fieldname in zone.get_referenced_object(
        #         'Zone_Conditioning_Equipment_List_Name').fieldnames if
        #     zone.get_referenced_object(
        #         'Zone_Conditioning_Equipment_List_Name
        #         ').get_referenced_object(
        #         fieldname) is not None
        # ]
        #     for zone in zone.idf.idfobjects[
        #         'ZoneHVAC:EquipmentConnections'.upper()]
        # }

        outs = energy_out.groupby("KeyValue").Value.sum()
        ins = energy_in.Value.sum()

        cop = float_round(outs.sum() / ins, 3)

        return cop

    def _get_setpoint_and_scheds(self, zone):
        """Gets temperature set points from sql EnergyPlus output and associated
        schedules.

        Args:
            zone (archetypal.template.zone.Zone): zone to gets information from
        """
        # Load the ReportData and filter *variable_output_name* and group by
        # zone name (*KeyValue*). Return annual average.
        variable_output_name = "Zone Thermostat Heating Setpoint Temperature"
        h_array = (
            ReportData.from_sql_dict(zone.sql)
            .filter_report_data(name=variable_output_name, keyvalue=zone.Name.upper())
            .loc[:, ["TimeIndex", "Value"]]
            .set_index("TimeIndex")
            .Value.values
        )

        heating_sched = UmiSchedule.from_values(
            Name=zone.Name + "_Heating_Schedule",
            values=(h_array > 0).astype(int),
            schTypeLimitsName="Fraction",
            idf=zone.idf,
        )

        variable_output_name = "Zone Thermostat Cooling Setpoint Temperature"
        c_array = (
            ReportData.from_sql_dict(zone.sql)
            .filter_report_data(name=variable_output_name, keyvalue=zone.Name.upper())
            .loc[:, ["TimeIndex", "Value"]]
            .set_index("TimeIndex")
            .Value.values
        )
        cooling_sched = UmiSchedule.from_values(
            Name=zone.Name + "_Cooling_Schedule",
            values=(c_array > 0).astype(int),
            schTypeLimitsName="Fraction",
            idf=zone.idf,
        )
        if np.all(c_array == 0):
            c_mean = np.NaN
        else:
            c_mean = c_array.mean()
        if np.all(h_array == 0):
            h_mean = np.NaN
        else:
            h_mean = h_array.mean()
        return h_mean, heating_sched, c_mean, cooling_sched

    def combine(self, other, weights=None):
        """Combine two ZoneConditioning objects together.

        Args:
            other (ZoneConditioning): The other ZoneConditioning object to
                combine with.
            weights (list-like, optional): A list-like object of len 2. If None,
                the volume of the zones for which self and other belongs is
                used.
        Returns:
            (ZoneConditioning): the combined ZoneConditioning object.
        """
        # Check if other is the same type as self
        if not isinstance(other, self.__class__):
            msg = "Cannot combine %s with %s" % (
                self.__class__.__name__,
                other.__class__.__name__,
            )
            raise NotImplementedError(msg)

        # Check if other is not the same as self
        if self == other:
            return self

        meta = self._get_predecessors_meta(other)

        if not weights:
            zone_weight = settings.zone_weight
            weights = [
                getattr(self._belongs_to_zone, str(zone_weight)),
                getattr(other._belongs_to_zone, str(zone_weight)),
            ]
            log(
                'using zone {} "{}" as weighting factor in "{}" '
                "combine.".format(
                    zone_weight,
                    " & ".join(list(map(str, map(int, weights)))),
                    self.__class__.__name__,
                )
            )

        a = self._float_mean(other, "CoolingCoeffOfPerf", weights)
        b = self._str_mean(other, "CoolingLimitType")
        c = self._float_mean(other, "CoolingSetpoint", weights)
        d = self._str_mean(other, "EconomizerType")
        e = self._float_mean(other, "HeatRecoveryEfficiencyLatent", weights)
        f = self._float_mean(other, "HeatRecoveryEfficiencySensible", weights)
        g = self._str_mean(other, "HeatRecoveryType")
        h = self._float_mean(other, "HeatingCoeffOfPerf", weights)
        i = self._str_mean(other, "HeatingLimitType")
        j = self._float_mean(other, "HeatingSetpoint", weights)
        k = any((self.IsCoolingOn, other.IsCoolingOn))
        l = any((self.IsHeatingOn, other.IsHeatingOn))
        m = any((self.IsMechVentOn, other.IsMechVentOn))
        n = self._float_mean(other, "MaxCoolFlow", weights)
        o = self._float_mean(other, "MaxCoolingCapacity", weights)
        p = self._float_mean(other, "MaxHeatFlow", weights)
        q = self._float_mean(other, "MaxHeatingCapacity", weights)
        r = self._float_mean(other, "MinFreshAirPerArea", weights)
        s = self._float_mean(other, "MinFreshAirPerPerson", weights)
        t = self.HeatingSchedule.combine(other.HeatingSchedule, weights)
        u = self.CoolingSchedule.combine(other.CoolingSchedule, weights)
        v = self.MechVentSchedule.combine(other.MechVentSchedule, weights)

        # create a new object with the previous attributes
        new_obj = self.__class__(
            **meta,
            CoolingCoeffOfPerf=a,
            CoolingLimitType=b,
            CoolingSetpoint=c,
            EconomizerType=d,
            HeatRecoveryEfficiencyLatent=e,
            HeatRecoveryEfficiencySensible=f,
            HeatRecoveryType=g,
            HeatingCoeffOfPerf=h,
            HeatingLimitType=i,
            HeatingSetpoint=j,
            IsCoolingOn=k,
            IsHeatingOn=l,
            IsMechVentOn=m,
            MaxCoolFlow=n,
            MaxCoolingCapacity=o,
            MaxHeatFlow=p,
            MaxHeatingCapacity=q,
            MinFreshAirPerArea=r,
            MinFreshAirPerPerson=s,
            HeatingSchedule=t,
            CoolingSchedule=u,
            MechVentSchedule=v,
        )
        new_obj._predecessors.extend(self.predecessors + other.predecessors)
        return new_obj



import functools
import logging as lg
import time
from sqlite3 import OperationalError

import numpy as np
import pandas as pd
from path import Path

from archetypal import log, EnergySeries


class ReportData(pd.DataFrame):
    """This class serves as a subclass of a pandas DataFrame allowing to add
    additional functionnality"""

    ARCHETYPE = "Archetype"
    REPORTDATAINDEX = "ReportDataIndex"
    TIMEINDEX = "TimeIndex"
    REPORTDATADICTIONARYINDEX = "ReportDataDictionaryIndex"
    VALUE = "Value"
    ISMETER = "IsMeter"
    TYPE = "Type"
    INDEXGROUP = "IndexGroup"
    TIMESTEPTYPE = "TimestepType"
    KEYVALUE = "KeyValue"
    NAME = "Name"
    REPORTINGFREQUENCY = "ReportingFrequency"
    SCHEDULENAME = "ScheduleName"
    UNITS = "Units"

    @classmethod
    def from_sql_dict(cls, sql_dict):
        report_data = sql_dict["ReportData"]
        report_data["ReportDataDictionaryIndex"] = pd.to_numeric(
            report_data["ReportDataDictionaryIndex"]
        )

        report_data_dict = sql_dict["ReportDataDictionary"]

        return cls(
            report_data.reset_index().join(
                report_data_dict, on=["ReportDataDictionaryIndex"]
            )
        )

    @classmethod
    def from_sqlite(
        cls,
        sqlite_file,
        table_name="WaterSystems:EnergyTransfer",
        warmup_flag=0,
        environment_type=3,
    ):
        """Reads an EnergyPlus eplusout.sql file and returns a :class:`ReportData`
        which is a subclass of :class:`DataFrame`.

        Args:
            environment_type (str): An enumeration of the environment type. (1 = Design
                Day, 2 = Design Run Period, 3 = Weather Run Period) See the various
                SizingPeriod objects and the RunPeriod object for details.
            sqlite_file (str):

        Returns:
            (ReportData): The ReportData object.
        """
        if not isinstance(sqlite_file, str):
            raise TypeError("Please provide a str, not a {}".format(type(sqlite_file)))
        file = Path(sqlite_file)
        if not file.exists():
            raise FileNotFoundError("Could not find sql file {}".format(file.relpath()))

        import sqlite3

        # create database connection with sqlite3
        with sqlite3.connect(sqlite_file) as conn:
            # empty dict to hold all DataFrames
            all_tables = {}
            # Iterate over all tables in the report_tables list
            sql_query = f"""
            SELECT rd.ReportDataIndex,
                   rd.TimeIndex,
                   rd.ReportDataDictionaryIndex,
                   red.ReportExtendedDataIndex,
                   t.Month,
                   t.Day,
                   t.Hour,
                   t.Minute,
                   t.Dst,
                   t.Interval,
                   t.IntervalType,
                   t.SimulationDays,
                   t.DayType,
                   t.EnvironmentPeriodIndex,
                   t.WarmupFlag,
                   p.EnvironmentType,
                   rd.Value,
                   rdd.IsMeter,
                   rdd.Type,
                   rdd.IndexGroup,
                   rdd.TimestepType,
                   rdd.KeyValue,
                   rdd.Name,
                   rdd.ReportingFrequency,
                   rdd.ScheduleName,
                   rdd.Units
            FROM ReportData As rd
                    INNER JOIN ReportDataDictionary As rdd ON rd.ReportDataDictionaryIndex = rdd.ReportDataDictionaryIndex
                    LEFT OUTER JOIN ReportExtendedData As red ON rd.ReportDataIndex = red.ReportDataIndex
                    INNER JOIN Time As t ON rd.TimeIndex = t.TimeIndex
                    JOIN EnvironmentPeriods as p ON t.EnvironmentPeriodIndex = p.EnvironmentPeriodIndex
            WHERE (IFNULL(t.WarmupFlag, 0) = @warmup_flag);
            """
            params = {"warmup_flag": warmup_flag}
            if table_name:
                conditions, table_name = cls.multiple_conditions(
                    "table_name", table_name, "Name"
                )
                sql_query = sql_query.replace(";", """ AND (%s);""" % conditions)
                params.update(table_name)
            if environment_type:
                conditions, env_name = cls.multiple_conditions(
                    "env_name", environment_type, "EnvironmentType"
                )
                sql_query = sql_query.replace(";", """ AND (%s);""" % conditions)
                params.update(env_name)
            df = cls.execute(conn, sql_query, params)
            return cls(df)

    @classmethod
    def multiple_conditions(cls, basename, cond_names, var_name):
        if not isinstance(cond_names, (list, tuple)):
            cond_names = [cond_names]
        cond_names = set(cond_names)
        cond_names = {
            "%s_%s" % (basename, i): name for i, name in enumerate(cond_names)
        }
        conditions = " OR ".join(
            ["%s = @%s" % (var_name, cond_name) for cond_name in cond_names]
        )
        return conditions, cond_names

    @staticmethod
    def execute(conn, sql_query, params):
        try:
            # Try regular str read, could fail if wrong encoding
            conn.text_factory = str
            df = pd.read_sql_query(sql_query, conn, params=params, coerce_float=True)
        except OperationalError as e:
            # Wring encoding found, the load bytes and decode object
            # columns only
            raise e
        return df

    @property
    def _constructor(self):
        return ReportData

    @property
    def schedules(self):
        return self.sorted_values(key_value="Schedule Value")

    @property
    def df(self):
        """Returns the DataFrame of the ReportData"""
        return pd.DataFrame(self)

    def heating_load(
        self, normalize=False, sort=False, ascending=False, concurrent_sort=False
    ):
        """Returns the aggragated 'Heating:Electricity', 'Heating:Gas' and
        'Heating:DistrictHeating' of each archetype

        Args:
            normalize (bool): if True, returns a normalize Series.
                Normalization is done with respect to each Archetype
            sort (bool): if True, sorts the values. Usefull when a load
                duration curve is needed.
            ascending (bool): if True, sorts value in ascending order. If a
                Load Duration Curve is needed, use ascending=False.

        Returns:
            EnergySeries: the Value series of the Heating Load with a Archetype,
                TimeIndex as MultiIndex.
        """
        hl = self.filter_report_data(
            name=("Heating:Electricity", "Heating:Gas", "Heating:DistrictHeating")
        )
        freq = list(set(hl.ReportingFrequency))
        units = list(set(hl.Units))
        freq_map = dict(Hourly="H", Daily="D", Monthly="M")
        if len(units) > 1:
            raise MixedUnitsError()

        hl = hl.groupby(["Archetype", "TimeIndex"]).Value.sum()
        log("Returned Heating Load in units of {}".format(str(units)), lg.DEBUG)
        return EnergySeries(
            hl,
            frequency=freq_map[freq[0]],
            units=units[0],
            normalize=normalize,
            sort_values=sort,
            ascending=ascending,
            to_units="kWh",
            concurrent_sort=concurrent_sort,
        )

    def filter_report_data(
        self,
        archetype=None,
        reportdataindex=None,
        timeindex=None,
        reportdatadictionaryindex=None,
        value=None,
        ismeter=None,
        type=None,
        indexgroup=None,
        timesteptype=None,
        keyvalue=None,
        name=None,
        reportingfrequency=None,
        schedulename=None,
        units=None,
        inplace=False,
    ):
        """filter RaportData using specific keywords. Each keywords can be a
        tuple of strings (str1, str2, str3) which will return the logical_or
        on the specific column.

        Args:
            archetype (str or tuple):
            reportdataindex (str or tuple):
            timeindex (str or tuple):
            reportdatadictionaryindex (str or tuple):
            value (str or tuple):
            ismeter (str or tuple):
            type (str or tuple):
            indexgroup (str or tuple):
            timesteptype (str or tuple):
            keyvalue (str or tuple):
            name (str or tuple):
            reportingfrequency (str or tuple):
            schedulename (str or tuple):
            units (str or tuple):
            inplace (str or tuple):

        Returns:
            pandas.DataFrame
        """
        start_time = time.time()
        c_n = []

        if archetype:
            c_1 = (
                conjunction(
                    *[self[self.ARCHETYPE] == archetype for archetype in archetype],
                    logical=np.logical_or,
                )
                if isinstance(archetype, tuple)
                else self[self.ARCHETYPE] == archetype
            )
            c_n.append(c_1)
        if reportdataindex:
            c_2 = (
                conjunction(
                    *[
                        self[self.REPORTDATAINDEX] == reportdataindex
                        for reportdataindex in reportdataindex
                    ],
                    logical=np.logical_or,
                )
                if isinstance(reportdataindex, tuple)
                else self[self.REPORTDATAINDEX] == reportdataindex
            )
            c_n.append(c_2)
        if timeindex:
            c_3 = (
                conjunction(
                    *[self[self.TIMEINDEX] == timeindex for timeindex in timeindex],
                    logical=np.logical_or,
                )
                if isinstance(timeindex, tuple)
                else self[self.TIMEINDEX] == timeindex
            )
            c_n.append(c_3)
        if reportdatadictionaryindex:
            c_4 = (
                conjunction(
                    *[
                        self[self.REPORTDATADICTIONARYINDEX]
                        == reportdatadictionaryindex
                        for reportdatadictionaryindex in reportdatadictionaryindex
                    ],
                    logical=np.logical_or,
                )
                if isinstance(reportdatadictionaryindex, tuple)
                else self[self.REPORTDATADICTIONARYINDEX] == reportdatadictionaryindex
            )
            c_n.append(c_4)
        if value:
            c_5 = (
                conjunction(
                    *[self[self.VALUE] == value for value in value],
                    logical=np.logical_or,
                )
                if isinstance(value, tuple)
                else self[self.VALUE] == value
            )
            c_n.append(c_5)
        if ismeter:
            c_6 = (
                conjunction(
                    *[self[self.ISMETER] == ismeter for ismeter in ismeter],
                    logical=np.logical_or,
                )
                if isinstance(ismeter, tuple)
                else self[self.ISMETER] == ismeter
            )
            c_n.append(c_6)
        if type:
            c_7 = (
                conjunction(
                    *[self[self.TYPE] == type for type in type], logical=np.logical_or
                )
                if isinstance(type, tuple)
                else self[self.TYPE] == type
            )
            c_n.append(c_7)
        if indexgroup:
            c_8 = (
                conjunction(
                    *[self[self.INDEXGROUP] == indexgroup for indexgroup in indexgroup],
                    logical=np.logical_or,
                )
                if isinstance(indexgroup, tuple)
                else self[self.INDEXGROUP] == indexgroup
            )
            c_n.append(c_8)
        if timesteptype:
            c_9 = (
                conjunction(
                    *[
                        self[self.TIMESTEPTYPE] == timesteptype
                        for timesteptype in timesteptype
                    ],
                    logical=np.logical_or,
                )
                if isinstance(timesteptype, tuple)
                else self[self.TIMESTEPTYPE] == timesteptype
            )
            c_n.append(c_9)
        if keyvalue:
            c_10 = (
                conjunction(
                    *[self[self.KEYVALUE] == keyvalue for keyvalue in keyvalue],
                    logical=np.logical_or,
                )
                if isinstance(keyvalue, tuple)
                else self[self.KEYVALUE] == keyvalue
            )
            c_n.append(c_10)
        if name:
            c_11 = (
                conjunction(
                    *[self[self.NAME] == name for name in name], logical=np.logical_or
                )
                if isinstance(name, tuple)
                else self[self.NAME] == name
            )
            c_n.append(c_11)
        if reportingfrequency:
            c_12 = (
                conjunction(
                    *[
                        self[self.REPORTINGFREQUENCY] == reportingfrequency
                        for reportingfrequency in reportingfrequency
                    ],
                    logical=np.logical_or,
                )
                if isinstance(reportingfrequency, tuple)
                else self[self.REPORTINGFREQUENCY] == reportingfrequency
            )
            c_n.append(c_12)
        if schedulename:
            c_13 = (
                conjunction(
                    *[
                        self[self.SCHEDULENAME] == schedulename
                        for schedulename in schedulename
                    ],
                    logical=np.logical_or,
                )
                if isinstance(schedulename, tuple)
                else self[self.SCHEDULENAME] == schedulename
            )
            c_n.append(c_13)
        if units:
            c_14 = (
                conjunction(
                    *[self[self.UNITS] == units for units in units],
                    logical=np.logical_or,
                )
                if isinstance(units, tuple)
                else self[self.UNITS] == units
            )
            c_n.append(c_14)

        filtered_df = self.loc[conjunction(*c_n, logical=np.logical_and)]
        log("filtered ReportData in {:,.2f} seconds".format(time.time() - start_time))
        if inplace:
            return filtered_df._update_inplace(filtered_df)
        else:
            return filtered_df.__finalize__(self)

    def sorted_values(self, key_value=None, name=None, by="TimeIndex", ascending=True):
        """Returns sorted values by filtering key_value and name

        Args:
            self: The ReporatData DataFrame
            key_value (str): key_value column filter
            name (str): name column filter
            by (str): sorting by this column name
            ascending (bool):

        Returns:
            ReportData
        """
        if key_value and name:
            return (
                self.filter_report_data(name=name, keyvalue=key_value)
                .sort_values(by=by, ascending=ascending)
                .reset_index(drop=True)
                .rename_axis("TimeStep")
                .set_index(["Archetype"], append=True)
                .swaplevel(i=-2, j=-1, axis=0)
            )
        else:
            return self.sort_values(by=by, inplace=False)


def conjunction(*conditions, logical=np.logical_and):
    """Applies a logical function on n conditons"""
    return functools.reduce(logical, conditions)


def or_conjunction(*conditions):
    return functools.reduce(np.logical_or, conditions)
